import os
import logging
import warnings
import yaml

logger = logging.getLogger(__name__)

CONFIG_FILE = os.path.join(os.environ["HOME"], ".config", "idoc", "idoc.yaml")
LOGGING_FILE = os.path.join(os.environ["HOME"], ".config", "idoc", "logging.yaml")
MACHINE_NAME=os.path.join(os.environ["HOME"], ".config", "idoc", "machine-name")
MACHINE_ID= "/etc/machine-id"


if not os.path.exists(os.path.dirname(CONFIG_FILE)):
    os.makedirs(os.path.dirname(CONFIG_FILE), exist_ok=True)

class IDOCConfiguration(object):
    """
    Handles the idoc configuration parameters
    Data are stored in and retrieved from a configuration file
    """

    _settings = {

        'default_class': {
            'board': "ArduinoDummy",
            'camera': "OpenCVCamera",
            'drawer': "DefaultDrawer",
            'result_writer': "CSVResultWriter",
            'roi_builder': "IDOCROIBuilder",
            'tracker': "AdaptiveBGModel"
        },

        'core' : {
            'debug': True
        },

        'network': {

            'port': 9000
        },

        'folders' : {
                        'results' : {'path' : '/idoc_data/results', 'description' : 'Where data will be saved by the saver class.'},
                        'paradigms': {'path': '/idoc_data/paradigms', 'description' : 'Where csv files containing hardware programs are stored.'}
        },

        # 'users' : {
        #               'admin' : { # pylint: disable=bad-continuation
        #                   'id' : 1, 'name' : 'admin', 'fullname' : '', 'PIN' : 9999, 'email' : '',
        #                   'telephone' : '', 'group': '', 'active' : False, 'isAdmin' : True,
        #                   'created' : datetime.datetime.now().timestamp()
        #               }
        # },

        'roi_builder': {
            'args': [],
            'kwargs': {},
            "target_coord_file": None,
        },

        'io': {
            'result_writer': {
                'args': [],
                'kwargs': {
                    'max_n_rows_to_insert': 1,
                }
            },

            'camera': {
                'args': [],
                'kwargs': {
                    'drop_each': 1,
                    'use_wall_clock': True,
                    'rotation': 0, #degs
                    'video_path': None
                }
            }
        },

        'controller': {
            'mapping_path': '/idoc_data/mega.csv',
            'paradigm_path': 'warm_up.csv',
            'arduino_port': "/dev/ttyACM0",
            'pwm': {
            },
        },

        'drawer': {
            'args': [],
            'kwargs': {
                'draw_frames': False,
                'video_out_fourcc': "DIVX",
                'framerate': None # match that of the camera
            },
            'last_drawn_path': "/tmp/last_img.png",
            'last_annot_path': "/tmp/last_img_annot.png",
        },

        'experiment': {
            'adaptation_time': 5,
            "max_duration": 18000,
            "location": None
        },
        "logging": LOGGING_FILE
    }


    def __init__(self, config_file=CONFIG_FILE, warn_if_exists=False):
        self._config_file = config_file
        self._warn_if_exists = warn_if_exists
        self.load()

    @property
    def content(self):
        return self._settings

    @property
    def file_exists(self):

        return os.path.exists(self._config_file)

    def save(self):
        """
        Save settings to settings file
        """

        try:
            with open(self._config_file, 'w') as filehandle:
                yaml.dump(self._settings, filehandle)

            logging.info('Saved idoc configuration file to %s', self._config_file)

        except:
            raise ValueError('Problem writing to file % s' % self._config_file)

    def load(self):
        """
        Reads saved configuration folders settings from configuration file
        If file does not exist, creates default settings
        """

        if self.file_exists:
            if self._warn_if_exists:
                logger.warning(f"{self._config_file} exists already")
            try:
                with open(self._config_file, 'r') as filehandle:
                    self._settings.update(yaml.load(filehandle, yaml.SafeLoader))
            except:
                raise ValueError("File %s is not a valid configuration file" % self._config_file)     

        else:
            self.save()


        return self._settings


def integrity_check():

    try:
        with open(CONFIG_FILE, "r") as filehandle:
            config = yaml.load(filehandle, yaml.SafeLoader)
            print(f"Can load {CONFIG_FILE}")
    except Exception as error:
        warnings.warn(error)
        print(f"Cannot load {CONFIG_FILE}")
    


if __name__ == "__main__":
    config = IDOCConfiguration()
    config.save()



import subprocess
import os.path

class RSession():

    def __init__(self, experiment_folder=None, decision_zone_mm=10, min_exits_required=5, max_time_minutes=60):

        self.experiment_folder = experiment_folder
        self.decision_zone_mm = decision_zone_mm
        self.min_exits_required = min_exits_required
        self.max_time_minutes = max_time_minutes

    def run(self):
        subprocess.call(['nautilus', self.experiment_folder])

        cmd = ['R', '--slave', '-e', 'lemdt_analysis <- LeMDTr::BehaviorAnalyzer$new( \
            experiment_folder = "{}", \
            decision_zone_mm = {}, \
            min_exits_required = {}, \
            max_time_minutes = {} \
            )\nlemdt_analysis$run()'.format(self.experiment_folder, self.decision_zone_mm, self.min_exits_required, self.max_time_minutes)]
        
        readable_cmd = ' '.join(cmd[3:])
        readable_cmd = ' '.join(readable_cmd.split())

        handle = open(os.path.join(self.experiment_folder, "script.R"), 'w')
        handle.write(readable_cmd + '\n')
        handle.close()
        subprocess.call(cmd)





from functools import wraps
import logging
import time
import traceback

from idoc.helpers import get_machine_id


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

id = get_machine_id()

def mixedomatic(cls):
    """ Mixed-in class decorator. """
    classinit = cls.__dict__.get('__init__')  # Possibly None.
    bases = [e for e in cls.__bases__ if e.__name__[-5:] == "Mixin"]

    # Define an __init__ function for the class.
    def __init__(self, *args, **kwargs):
        # Call the __init__ functions of all the bases.
        for base in bases:
            base.__init__(self, *args, **kwargs)
        # Also call any __init__ function that was in the class.
        if classinit:
            classinit(self, *args, **kwargs)

        for base in bases:
            __init2__ = base.__dict__.get('__init2__')
            if __init2__:
                __init2__(self, *args, **kwargs)

    # Make the local function the class's __init__.
    setattr(cls, '__init__', __init__)
    return cls


class WrongMachineID(Exception):
    r"""
    An exception to be raised when the id
    parsed from the requested URL
    does not match that of the local machine
    """
    pass

def wrong_id(f):
    def wrapper(**kwargs):
        if kwargs["id"] != id:
            raise WrongMachineID
        else:
            if "id" in f.__code__.co_varnames:
                return f(**kwargs)
            else:
                kwargs.pop('id')
                return f(**kwargs)

    return wrapper


def error_decorator(func):
    """
    A simple decorator to return an error dict so we can display it in the webUI
    """
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logging.error(traceback.format_exc())
            return {'error': traceback.format_exc()}
    return func_wrapper

def warning_decorator(func):
    """
    A simple decorator to return an error dict so we can display it in the webUI
    Less verbose than error
    """
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logging.error(traceback.format_exc())
            return {'error': str(e)}
    return func_wrapper


def retry(ExceptionToCheck, tries=1, delay=1, backoff=2):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry
    """
    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)
        return f_retry
    return deco_retry


# https://stackoverflow.com/a/4761058
import logging
from pathlib import Path
import sys
import unittest

from . import test_generic
from LeMDT.LeMDT.interface import Interface
from LeMDT.LeMDT.interface import Tracker
from pypylon import pylon, genicam


class TestCamera(unittest.TestCase, Interface):

    def __init__(self, *args, **kwargs):

        interface = Interface(
             arduino = False, track = True,
             camera = 'pylon', duration = 10
        )
        
        self.tracker = Tracker(interface=interface, camera=interface.camera)
        unittest.TestCase.__init__(self, *args, **kwargs)

    def test_camera_can_be_loaded(self):
        """
        Make sure that the camera can be loaded
        """
        self.tracker.load_camera()
    
    def test_camera_can_read_frames(self):

        self.tracker.load_camera()
        self.tracker.stream.read_frame()



# for now it will not scan devices, just retrieve attributes of a device running on localhost
import json
import logging
from threading import Thread
import time
import urllib.error
import urllib.request
import urllib.parse

import requests

from idoc.client.utils.logging_server import LogRecordSocketReceiver
from idoc.client.utils.mixins import HTTPMixin
from idoc.debug import IDOCException, ScanException

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



class Device(Thread, HTTPMixin):
    r"""
    Individual Drosophila Optogenetic Conditioner Device
    """
    _remote_pages = {
            'id' : "id",
            # 'videofiles' :  "data/listfiles/video",
            # 'stream' : "stream.mjpg",
            # 'user_options' : "user_options",
            'log' : "data/log",
            'static' : "static",
            'controls' : "controls",
            'logs' : 'logs'
            # 'machine_info' : "machine",
            # 'update' : "update"
            }



    _allowed_instructions_status = { "stream": ["stopped"],
                                     "start": ["stopped"],
                                     "start_record": ["stopped"],
                                     "stop": ["streaming", "running", "recording"],
                                     "poweroff": ["stopped"],
                                     "reboot" : ["stopped"],
                                     "restart" : ["stopped"],
                                     "offline": []}




    def __init__(self, ip, port, *args, remote=False, log_server=False, **kwargs):

        self._ip = ip
        self._port = port
        self._info = {}
        self._settings = {}
        self._id_url = "http://%s:%i/%s" % (ip, port, self._remote_pages['id'])
        self._id = ""
        self._log_server = log_server
        self._remote = remote

        self.running = False
        self.stopped = False

        if self._log_server:
            self._logs_cache = []
            self._tcpserver = LogRecordSocketReceiver()

        super().__init__(*args, **kwargs)

    def get_api(self, api_endpoint, remote=False):

        if api_endpoint is list:
            api_endpoint = "/".join(api_endpoint)


        if self._remote or remote:
            url = "http://%s:%s/device/%s/%s" % (
                self._ip, self._port,
                self.id, api_endpoint
            )

        else:
            url = "http://%s:%s/%s/%s" % (
                self._ip, self._port,
                api_endpoint, self.id
            )

        return url


    def controls(self, submodule, action):

        url = self.get_api([self._remote_pages["controls"], submodule, action])
        return self._get_json(url)

    def close(self, answer="Y"):

        url = self.get_api("close")
        self._get_json(url, post_data={"answer": answer})

    def get_paradigms(self):

        url = self.get_api("list_paradigms")
        return self._get_json(url)

    def post_paradigm(self, paradigm_path):
        """
        :param paradigm_path: A path to a paradigm.csv that IDOC controllers understand
        :type paradigm_path: bytes
        """

        url = self.get_api("load_paradigm")
        requests.post(url, data=paradigm_path)
        # self._get_json(url, post_data={"paradigm_path": paradigm_path})

    def post_settings(self, settings):
        """
        Post a settings dictionary
        """
        url = self.get_api("settings")
        self._get_json(url, post_data=json.dumps(settings).encode())


    @property
    def settings(self):
        return self._settings

    @settings.getter
    def settings(self):

        url = self.get_api("settings")
        self._settings = self._get_json(url)
        return self._settings

    def start_experiment(self):
        url = self.get_api("controls/control_thread/start_experiment")
        self._get_json(url)

    def stop_experiment(self):
        url = self.get_api("controls/control_thread/stop_experiment")
        self._get_json(url)

    def reset_experiment(self):
        url = self.get_api("controls/control_thread/reset_experiment")
        self._get_json(url)

    def warm_up(self):
        url = self.get_api("controls/control_thread/warm_up")
        self._get_json(url)

    @property
    def info(self):
        return self._info

    def get_info(self):
        logging.debug('Getting self._info')
        url = self.get_api("info")
        self._info = self._get_json(url)
        logging.debug('_get_json self._info')
        return self._info

    # def get_logs(self):
    #     return self._logs_cache

    # def post_logs(self, post_data):
    #     parsed_data = urllib.parse.parse_qs(post_data.decode())
    #     # print(parsed_data)
    #     self._logs_cache.extend(parsed_data["logs"])


    def _reset_info(self):
        r"""
        This is called whenever the device goes offline or when it is first added
        """
        self._info = {'status': "offline"}

    @property
    def id(self):
        return self._id

    @id.getter
    def id(self):
        r"""
        """
        old_id = self._id
        try:
            resp = self._get_json(self._id_url)
        except ScanException:
            return self._id

        self._id = resp['id']
        if self._id != old_id:
            if old_id:
                logging.warning("Device id changed at %s. %s ===> %s", self._ip, old_id, self._id)
            self._reset_info()

        # keep the ip
        self._info["ip"] = self._ip
        return self._id

    def ip(self):
        return self._ip


    def run(self):

        self.running = True

        if self._log_server:
            print('About to start TCP server...')
            self._tcpserver.serve_until_stopped()

        while not self.stopped:
            # Implement here whatever I need
            time.sleep(1)

    def stop(self):

        self.stopped = True
        if self._log_server:
            self._tcpserver.abort = True

    def last_image(self):
        """
        Collects the last drawn image fromt the device
        TODO: on the device side, this should not rely on an actuale image file but be fished from memory
        """
        # we return none if the device is not in a stoppable status (e.g. running, recording)
        if self.info["status"] not in self._allowed_instructions_status["stop"]:
            return None
        try:
            img_path = self.info["recognizer"]["last_annot_path"].lstrip("/")
        except KeyError:
            raise KeyError("Cannot find last image for device %s" % self._id)

        img_url = self.get_api([self._remote_pages['static'], img_path], remote=False)
        # img_url = "http://%s:%i/%s/%s" % (self._ip, self._port, self._remote_pages['static'], img_path)
        try:
            return urllib.request.urlopen(img_url, timeout=5)
        except urllib.error.HTTPError:

            raise Exception("Could not get image for ip = %s (id = %s)" % (self._ip, self._id))


class DeviceScanner(Thread):

    def __init__(self, port, *args, device_refresh_period=5, deviceClass=Device, remote=False, **kwargs):

        self.devices = []
        self._port = port
        self._device_refresh_period = device_refresh_period
        self._deviceClass = deviceClass
        self._remote = remote
        self._max_trials = 50
        self.running = False
        self.stopped = False
        super(DeviceScanner, self).__init__(*args, **kwargs)


    def get_all_devices_info(self):

        '''
        Returns a dictionary in which each entry has key of device_id
        '''
        out = {}

        for device in self.devices:
            out[device.id] = device.info
        return out

    def get_device(self, idx):
        return self.devices[idx]

    def get_device_by_id(self, machine_id, silent=True, trials=0):

        if silent and trials == 0:
            print("Connecting to IDOC device in machine %s" % (machine_id))

        for device in self.devices:
            try:
                info_id = device.info['id']
            except KeyError:
                continue

            if info_id == machine_id:
                return device

        if silent or trials > self._max_trials: # pylint: disable=no-else-return
            trials += 1
            time.sleep(2)
            return self.get_device_by_id(machine_id, trials=trials)

        else:
            message = "Device with machine_id %s is not available." % machine_id
            raise IDOCException(message)

    def run(self):

        self.running = True

        device = self._deviceClass(ip="localhost", port=self._port, remote=self._remote)
        device.start()
        self.devices.append(device)

        while not self.stopped:
            time.sleep(self._device_refresh_period)

            for device in self.devices:
                device.get_info()


    def stop(self):
        for device in self.devices:
            device.stop()

        self.stopped = True



import pickle
import logging
import logging.handlers
import socketserver
import struct
import urllib.request

import requests


from idoc.client.utils.mixins import HTTPMixin

class LogRecordStreamHandler(socketserver.StreamRequestHandler, HTTPMixin):
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def __init__(self, *args, **kwargs):
        self._cache = []
        super().__init__(*args, **kwargs)

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack('>L', chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.unPickle(chunk)
            record = logging.makeLogRecord(obj)

            peer_ip = self.connection.getpeername()[0]
            url = "http://%s:%d/%s" % (peer_ip, 9000, "id")
            resp = self._get_json(url)

            # req = urllib.request.Request(url)
            # file_handle = urllib.request.urlopen(req, timeout=5)
            # message = file_handle.read()
            # resp = json.loads(message)
            machine_id = resp["id"]
            self.handleLogRecord(record, machine_id)


    def unPickle(self, data):
        return pickle.loads(data)

    def handleLogRecord(self, record, machine_id):
        url = "http://%s:%d/device/%s/post_logs" % ("localhost", 8000, machine_id)

        requests.post(url, data={"logs": record.msg})
        # self._get_json(url, post_data={"logs": record.msg})

        if self.server.logname is not None:
            name = self.server.logname
        else:
            name = record.name



        logger = logging.getLogger(name)
        # N.B. EVERY record gets logged. This is because Logger.handle
        # is normally called AFTER logger-level filtering. If you want
        # to do filtering, do it at the client end to save wasting
        # cycles and network bandwidth!
        logger.handle(record)

class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    Simple TCP socket-based logging receiver suitable for testing.
    """

    allow_reuse_address = True

    def __init__(self, host='localhost',
                 port=logging.handlers.DEFAULT_TCP_LOGGING_PORT,
                 handler=LogRecordStreamHandler):

        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.abort = 0
        self.timeout = 1
        self.logname = None

    def serve_until_stopped(self):
        import select
        abort = 0
        while not abort:
            rd, wr, ex = select.select([self.socket.fileno()],
                                       [], [],
                                       self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort


def main():
    logging.basicConfig(
        format='%(relativeCreated)5d %(name)-15s %(levelname)-8s %(message)s')
    tcpserver = LogRecordSocketReceiver()
    print('About to start TCP server...')
    tcpserver.serve_until_stopped()

if __name__ == '__main__':
    main()



import numpy as np
import cv2

white = (255, 255, 255)
yellow = (0, 255, 255)
red = (0, 0, 255)
blue = (255, 0, 0)

cv2_version = cv2.__version__
 

class Arena():

    ## TODO
    ## Make config an argument that the user can change from the CLI
    def __init__(self, tracker, contour, config = "config.yml"):
        """Initialize an arena object
        """
        self.tracker = tracker
        self.contour = contour
        self.min_arena_area = self.tracker.interface.min_arena_area
        self.area = None
        self.corners = None
        self.column = None
        self.identity = None
        self.mask = None
        self.box = None
        self.fly = None
        self.tl_corner = None
        self.br_corner  = None
        self.width = self.tracker.interface.cfg['arena']['width']
        self.height = self.tracker.interface.cfg['arena']['height']
        
        self.dilation_kernel = np.ones((5,5),np.uint8)
        fly = None

    def __str__(self):
        return "Arena(tracker = self, contour = arena_contour, identity = {})".format(self.identity)
        
    def compute(self):
        # try:
        self.area = cv2.contourArea(self.contour)
       
        M0 = cv2.moments(self.contour)
        if M0['m00'] != 0 and M0['m10']:
            cx = int(M0['m10']/M0['m00'])
            cy = int(M0['m01']/M0['m00'])
            box = np.array([
                [cx-self.width, cy-self.height], [cx-self.width, cy+self.height],
                [cx+self.width, cy+self.height], [cx+self.width, cy-self.height]
            ])  
        
            
        else:
            rect = cv2.minAreaRect(self.contour)
            box = cv2.boxPoints(rect)
            box = np.int0(box)
            cx = np.mean([pt[0] for pt in box])
            cy = np.mean([pt[1] for pt in box])
        
        self.box = box
        center = (cx, cy)
        self.center = center
        self.cx  = cx
        self.cy = cy


            # handle what happens when the if above is not true
       

        tl_corner = np.min(box, axis = 0)
        br_corner = np.max(box, axis = 0)
    
        self.tl_corner = tl_corner 
        self.br_corner = br_corner 

        
        corners = np.array([tl_corner, br_corner])
        return corners

    def set_id(self, identity):
        self.identity = identity

    def set_column(self, column_number):
        self.column = column_number

    def validate(self):
        
        if self.area < self.min_arena_area * self.tracker.interface.fraction_area:            
            return False
        else:
            return True
        
    def make_mask(self, shape):

        mask = np.full(shape, 0, dtype = np.uint8)
        cv2.drawContours(mask, [self.box], 0, white, -1)
        cv2.dilate(mask, self.dilation_kernel, iterations = 1)
        self.mask = mask
        return mask

    def draw(self, img):
        """
        Draw the arena
        """
      
        self.tracker.log.debug('Arena {} has an area of {}'.format(self.identity, self.area))
        cv2.putText(img, str(self.identity), (self.tl_corner[0] - 50, self.tl_corner[1] - 25), cv2.FONT_HERSHEY_SIMPLEX, 1, (255,255,255), 2)

        # Draw the actual contour
        cv2.drawContours(img, [self.contour], 0, red, 1)
        #cv2.rectangle(img, (self.x, self.y), (self.x + self.w, self.y + self.h), (0,255,255), 2)

        # Draw the detected box
        cv2.drawContours(img,[self.box], 0, yellow, 2)
        #cv2.circle(img, (self.cx, self.cy), 2, blue, -1)
        
        return img
    
    def crop(self,img):
        return cv2.cvtColor(img[self.box[0,1]:self.box[1,1], self.box[0,0]:self.box[2,0]], cv2.COLOR_BGR2RGB)

    def find_flies(self, transform, kernel):
        #transform = cv2.morphologyEx(transform, cv2.MORPH_OPEN, kernel)

        if cv2_version[0] == '4':
            flies, _ = cv2.findContours(transform, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        else:
            _, flies, _ = cv2.findContours(transform, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

        if len(flies) == 0:
            pass
            # try the edge insensitive method

        return flies
 
    

class Fly():
    
    ## TODO
    ## Make config an argument that the user can change from the CLI
    def __init__(self, tracker, arena, contour, identity, config = "config.yml"):
        """Initialize fly object
        """
       
        self.tracker = tracker
        self.contour = contour
        self.identity = identity
        self.arena  = arena


        self.min_object_length = self.tracker.interface.cfg["fly"]["min_length"]
        self.min_obj_arena_dist =  self.tracker.interface.cfg["fly"]["min_dist_arena"]
        self.max_object_area =  self.tracker.interface.cfg["fly"]["max_area"]
        self.min_object_area =  self.tracker.interface.cfg["fly"]["min_area"]
        self.min_intensity =  self.tracker.interface.cfg["fly"]["min_intensity"]

        self.x = None
        self.y = None
        self.w = None
        self.h = None
        self.diagonal = None
        self.arena_distance_x = None
        self.arena_distance_y = None
    
    def compute(self):
        # compute area of the fly contour
        self.area = cv2.contourArea(self.contour)
        # compute the coordinates of the bounding box around the contour
        # https://docs.opencv.org/3.1.0/dd/d49/tutorial_py_contour_features.html
        self.x, self.y, self.w, self.h = cv2.boundingRect(self.contour)

        self.arena_distance_x = self.x - self.arena.tl_corner[0]
        self.arena_distance_y = self.y - self.arena.tl_corner[1]

        # compute the length of the diagonal line between opposite corners of the bounding box
        self.diagonal = np.sqrt(self.w ** 2 + self.h ** 2) 

        # compute the center of the fly
        M1 = cv2.moments(self.contour)
        if M1['m00'] != 0 and M1['m10']:
            # used in the validation step
            self.cx = int(M1['m10']/M1['m00'])
            self.cy = int(M1['m01']/M1['m00'])
            self.x_corrected = self.cx - self.arena.center[0]
            self.y_corrected = self.cy - self.arena.center[1]
        else:
            # handle what happens when the if above is not true
            pass
    
    def validate(self, gray):
        """Validate fly by checking all requirements
        Returns True if all requirements are passed and False otherwise
        """

        # the diagonal needs to be greater than a minimum
        fly_passed_1 = self.diagonal > self.min_object_length
        # the intensity on the top left corner needs to be greater than a minimum 
        ## INCONGRUENT
        intensity = gray[self.y, self.x]
        fly_passed_2 = intensity > self.min_intensity
        # the area has to be more than a minimum and less than a maximum
        fly_passed_3 = self.area > self.min_object_area
        fly_passed_4 = self.area < self.max_object_area

        ## DEBUG
        self.tracker.log.debug(self.arena.identity)
        self.tracker.log.debug('Diagonal test: {} --> {}'.format(self.diagonal, fly_passed_1))
        self.tracker.log.debug('Intensity test: {} --> {}'.format(intensity, fly_passed_2))
        self.tracker.log.debug('Min area test: {} --> {}'.format(self.area, fly_passed_3))
        self.tracker.log.debug('Max area test: {} --> {}'.format(self.area, fly_passed_4))

        if not (fly_passed_1 and fly_passed_2 and fly_passed_3 and fly_passed_4) or getattr(self, "cx") is None:
                return False
        else:
            # the center of the fly contour needs to be within the contour of the arena it was found in
            # it has to be more than min_obj_arena_dist pixels from the arena contour (inside it)
            # if min_obj_arena_dist is 0, then it's fine if it borders the arena contour
            fly_passed_5 = cv2.pointPolygonTest(self.arena.contour, (self.cx, self.cy), True) > self.min_obj_arena_dist
            return fly_passed_5 
        
    def draw(self, img, frame_count, relative_to_arena=False):
        """Draw fly
        """
        
        if self.identity > 1:
            self.tracker.log.debug('Error: more than one object found per arena')

        if relative_to_arena:
            cv2.rectangle(img, (self.arena_distance_x, self.arena_distance_y), (self.arena_distance_x + self.w, self.arena_distance_y + self.h), blue, 2)
        else:
            cv2.rectangle(img, (self.x, self.y), (self.x + self.w, self.y + self.h), blue, 2)
        self.tracker.log.debug("tl: {},{}. br: {},{}".format(self.x, self.y, self.x + self.w, self.y + self. h))

        return img

    def save(self, frame_count, time_position):
        record = "\t".join(["{}"] * 10) + "\n"
        record.format(frame_count, time_position, self.arena.identity, self.identity,
                      self.cx, self.cy, self.arena.cx, self.arena.cy, self.cx-self.arena.cx, self.cy-self.arena.cy)
        return record



# Standard library imports
import logging

# Third party imports
import cv2
import imutils
import numpy as np
from sklearn.cluster import KMeans

# Local application imports

from idoc.server.trackers.features import Arena, Fly
from idoc.server.io.cameras import CAMERAS
from idoc.configuration import IDOCConfiguration
from idoc.server.core.base import Base

# Set up package configurations

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

cv2_version = cv2.__version__ # pylint: disable=no-member

def crop_camera(img, crop):
    width = img.shape[1]
    fw = width//crop
    y0 = fw * (crop // 2)
    if len(img.shape) == 3:
        img = img[:, y0:(y0+fw), :]
    else:
        img = img[:, y0:(y0+fw)]
    return img


class CouldNotReadFrameError(Exception):
    """
    Used to signal that no more frames are coming from the camera
    """

class LessArenasThanTargetError(Exception):
    """
    Used to signal the target number of arenas (usually 20) was not detected in a frame
    """

class SimpleTracker(Base):

    _live_feed_path = '/tmp/last_img.jpg'
    row_all_flies = {}
    _info = {}

    frame_color = None
    frame_original = None
    _columns = ["frame_count", "arena", "x", "y"]

    def __init__(self, *args, camera_name="opencv", video_path=None, **kwargs):
        r"""
        Setup video recording parameters.

        :param camera_name: Codename of the camera selected by the user.
        Must match one of the keys in CAMERAS
        :type camera_name: str

        :param video: Path to video file for offline tracking. If None, live tracking is performed.
        :type video: str
        """
        super().__init__(*args, **kwargs)

        # Load a camera
        self._camera_name = camera_name
        self.video_path = video_path
        self.load_camera(self.video_path)

        # tracking framerate
        # if we are reading from a video
        # it will be limited by the processing capacity of the PC
        # if we are reading from a live feed
        # unless the feed has very high FPS,
        # the tracking framerate is given by the camera framerate
        self._framerate = 0

        # Config variables
        config = IDOCConfiguration()
        self.targets = config.content["arena"]["targets"]
        self.sampled_framerate = 0
        self.block_size = config.content["arena"]["block_size"]
        self.param1 = config.content["arena"]["param1"]
        self.crop = config.content["tracker"]["crop"]
        self.kernel_factor = config.content["arena"]["kernel_factor"]
        self.kernel = np.ones((self.kernel_factor, self.kernel_factor), np.uint8)
        self.N = config.content["tracker"]["N"]
        self._delta_t = 2

        # Results variables
        # how many flies are found in the whole experiment
        # in a perfect recording, it should be nframes x nflies/frame
        self.found_flies = 0
        # frames recorded since pressing play
        self.frame_count = 0
        # flies detected in the last frame
        self.old_found = 0
        # number of arenas where no fly was detected
        self.missing_fly = 0

        self.stacked_arenas = []
        self.arena_contours = []
        self.fraction_area = 1

        self.arenas = {}
        self.masks = {}


    @property
    def columns(self):
        return self._columns

    @property
    def camera(self):
        return self._submodules["camera"]

    @camera.setter
    def camera(self, camera):
        self._submodules["camera"] = camera
        self._settings.update(camera.settings)

    # @if_not_self_camera
    def load_camera(self, video_path=None):
        """
        Populate the camera attribute of the Tracker class
        Make it an instance of the classes in CAMERAS.py
        """

        config = IDOCConfiguration()
        self._CameraClass = CAMERAS[self._camera_name] # pylint: disable=invalid-name
        self.camera = self._CameraClass(video_path=video_path)
        # TODO What happens when we are reading a video?
        self.receive()

        logger.info("camera has shape w: %d h:%d", *self.camera.resolution)

    def init_image_arrays(self):
        """
        Initialize containers for the images that will be produced by the device
        These are required for the GUI to display something before start
        """

        # TODO Sort this mess

        # Make accu_image an array of size heightxwidthxnframes that will
        # store in the :, :, i element the result of the tracking for frame i
        self.accu_image = np.zeros((*self.camera.shape, self.N), np.uint8)

        # self.img = np.zeros((self.video_height, self.video_width, 3), np.uint8)
        self.transform = np.zeros(*self.camera.shape, np.uint8)
        self.gray_masked = np.zeros(self.camera.shape, np.uint8)
        self.frame_color = np.zeros(self.camera.shape, np.uint8)
        self.gray_color = np.zeros(self.camera.shape, np.uint8)
        self.gray_gui = np.zeros(self.camera.shape, np.uint8)

    def rotate_frame(self, img, rotation=180):
        # Rotate the image by certan angles, 0, 90, 180, 270, etc.
        rows, cols = img.shape[:2]
        M = cv2.getRotationMatrix2D((cols / 2, rows / 2), rotation, 1)
        img_rot = cv2.warpAffine(img, M, (cols, rows))
        return img_rot

    def annotate_frame(self, img):
        """
        Take a np.array and annotate it to provide info about frame number and time
        """

        ## TODO Dont make coordinates of text hardcoded
        ###################################################
        # cv2.putText(img, 'Frame: '+ str(self.record_frame_count), (25, 25), cv2.FONT_HERSHEY_SIMPLEX, 1, (40, 170, 0), 2)
        # cv2.putText(img, 'Time since tracker running: '+  str(int(self.timestamp_seconds)), (1000, 25), cv2.FONT_HERSHEY_SIMPLEX, 1, (40, 170, 0), 2)
        cv2.putText(img, 'LEFT', (25, 525), cv2.FONT_HERSHEY_SIMPLEX, 1, (40, 170, 0), 2)
        cv2.putText(img, 'RIGHT', (1100, 525), cv2.FONT_HERSHEY_SIMPLEX, 1, (40, 170, 0), 2)
        return img

    def find_arenas(self, gray, range_low_1=0, range_up_1=255):
        """
        Performs Otsu thresholding followed by morphological transformations to improve the signal/noise.
        Input image is the average image
        Output is a tuple of length 2, where every elemnt is either a list of contours or None (if no contours were found)

        More info on Otsu thresholding: https://docs.opencv.org/3.4/d7/d4d/tutorial_py_thresholding.html

        gray: Gray frame as output of read_frame()
        range_low_1: value assigned to pixels below the Otsu threshold
        range_up_1: value assigned to pixels above the Otsu threshold
        kernel_factor: size of the square kernel used in blurring and openings
        """

        # assign the grayscale to the ith frame in accu_image
        self.accu_image[:, :, self.frame_count % self.N] = gray
        if self.frame_count > 1:
            average = np.average(self.accu_image[:, :, :self.frame_count], 2).astype('uint8')
        else:
            average = self.accu_image[:, :, self.frame_count].astype('uint8')

        image_blur = cv2.GaussianBlur(average, (self.kernel_factor, self.kernel_factor), 0)
        _, thresholded = cv2.threshold(image_blur, range_low_1, range_up_1, cv2.THRESH_BINARY+cv2.THRESH_OTSU)
        #arena_opening = cv2.morphologyEx(image1, cv2.MORPH_OPEN, self.kernel, iterations = 2)

        if cv2_version[0] == '4':
            arena_contours, _ = cv2.findContours(thresholded, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        else:
            _, arena_contours, _ = cv2.findContours(thresholded, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)


        # to merge contours belonging to the same chamber
        # specially on ROI 10
        contour_mask = np.zeros_like(gray)

        for contour in arena_contours:
            cv2.fillPoly(contour_mask, pts=[contour], color=(255, 255, 255))



        dilation = cv2.dilate(contour_mask, (5, 5), 10)
        erosion = cv2.erode(dilation, (5, 5), 10)

        # cv2.imshow('thr', image1)
        # cv2.imshow('erosion', erosion)
        # cv2.imshow('contour mask', contour_mask)
        # if cv2.waitKey(33) == ord('q'):
        #     self.close()

        if cv2_version[0] == '4':
            arena_contours, _ = cv2.findContours(erosion, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        else:
            _, arena_contours, _ = cv2.findContours(erosion, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

        return arena_contours

    def track(self):
        # TODO  Either rewrite the whole thing
        # ditch for an already existing Tracker class
        # that we can integrate here

        config = IDOCConfiguration()
        targets = config.content["arena"]["targets"]

        if not self.stopped:

            # Read a new frame
            _, frame = self.camera.read_frame()
            frame = crop_camera(frame, self.crop)

            # Make single channel i.e. grayscale
            if len(frame.shape) == 3:
                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
                gray_color = frame
            else:
                gray = frame.copy()
                gray_color = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)

            # Annotate frame
            frame_color = self.annotate_frame(gray_color)


            # Find arenas for the first N frames
            if self.frame_count < self.N and self.frame_count > 0 or not self.ready:
                self.arena_contours = self.find_arenas(gray)

            if self.arena_contours is None:
                found_targets = 0
            else:
                found_targets = len(self.arena_contours)

            if found_targets < int(targets):
                self.frame_count += 1
                self.frame_color = gray_color
                raise LessArenasThanTargetError("Number of putative arenas found less than target. Discarding frame")


            if not self.ready:
                return True


            transform = cv2.adaptiveThreshold(gray, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, self.block_size, self.param1)
            self.transform = transform

            arenas_list = [None] * targets


            # Initialize an arena identity that will be increased with 1
            # for every validated arena i.e. not on every loop iteration!
            id_arena = 1
            # For every arena contour detected

            for arena_contour in self.arena_contours:

                # Initialize an Arena object
                arena = Arena(tracker = self, contour = arena_contour)
                # and compute parameters used in validation and drawing
                arena.corners = arena.compute()
                # Validate the arena
                # If not validated, move on the next arena contour
                # i.e. ignore the current arena
                if not arena.validate():
                    # move on to next i.e. continue to the next iteration
                    # continue means we actually WON'T continue with the current arena
                    continue

                if id_arena <= targets:
                    arenas_list[id_arena-1] = arena
                    # Update the id_arena to account for one more arena detected
                    id_arena += 1
                else:
                    break

            found_arenas = np.sum([a is not None for a in arenas_list])

            if found_arenas != self.targets:
                self.frame_count += 1
                raise LessArenasThanTargetError("Number of arenas found not equal to target. Discarding frame")


            columns = config.content['arena']['columns']
            x_coord = np.array([a.corners[1][0] for a in arenas_list])
            kmeans = KMeans(n_clusters=columns).fit(x_coord.reshape(len(x_coord), 1))

            cluster_mean_x = [0, ] * columns
            cluster_centers = list(range(columns))
            for i, col in enumerate(range(columns)):
                cluster_mean_x[i] = np.mean(x_coord[kmeans.labels_ == col])

            indices = sorted(range(len(cluster_mean_x)), key=lambda k: cluster_mean_x[k])
            labels = np.array(cluster_centers)[indices][kmeans.labels_]

            for i, arena in enumerate(arenas_list):
                arena.set_column(labels[i])


            # sort the arenas_list using the br corner attribute (.corners[1])
            # the corner is a tuple showing (x, y)
            # we want to sort first on y ([1]) and then on x ([0])
            # both in decreasing order because the
            sorted_arenas_br_to_tl_horizontally = sorted(arenas_list, key=lambda arena: (arena.column, arena.corners[1][1]))

            for identity, arena in enumerate(sorted_arenas_br_to_tl_horizontally):

                arena.set_id(identity+1)

                # If validated, keep analyzing the current arena
                frame_color = arena.draw(frame_color)

                # Make a mask for this arena
                ## TODO Right now the masking module does not work
                ## and is not required either
                mask = arena.make_mask(gray.shape[:2])

                ## DEBUG
                ## Add the arena to the dictionary
                self.arenas[arena.identity] = arena
                self.masks[arena.identity] = mask

                ## Find flies in this arena
                gray_masked = cv2.bitwise_and(self.transform, mask)
                fly_contours = arena.find_flies(gray_masked, self.kernel)

                # Initialize a fly identity that will be increased with 1
                # for every validated fly i.e. not on every loop iteration!
                id_fly = 1

                putative_flies = []

                # For every arena contour detected
                for fly_contour in fly_contours:
                    # Initialize a Fly object
                    fly = Fly(tracker=self, arena=arena, contour=fly_contour, identity=id_fly)
                    # and compute parameters used in validation and drawing
                    fly.compute()


                    # If not validated, move on the next fly contour
                    # i.e. ignore the current fly
                    if not fly.validate(gray):
                        logger.debug("Fly not validated")
                        continue

                    putative_flies.append(fly)

                if len(putative_flies) == 0:
                    self.missing_fly += 1
                    # fname = str(self.record_frame_count) + "_" + str(arena.identity) + ".tiff"
                    # gray_crop = gray[arena.tl_corner[1]:arena.br_corner[1], arena.tl_corner[0]:arena.br_corner[0]]
                    # cv2.imwrite(Path(self.failed_arena_path, fname).__str__(), gray_crop)
                    continue
                elif len(putative_flies) > 1:
                    logger.debug("Arena %d in frame #%d. Multiple flies found", arena.identity, self.frame_count)
                    fly = sorted(putative_flies, key=lambda f: f.area, reverse=True)[0]
                else:
                    fly = putative_flies[0]


                logger.debug("Fly %d in arena %d validated with area %3.f and length %3.f", fly.identity, fly.arena.identity, fly.area, fly.diagonal)

                row = {
                    "frame": self.frame_count,
                    "arena": arena.identity,
                    #"fly": fly.identity,
                    "x": fly.x_corrected,
                    "y": fly.y_corrected,
                    }

                self.row_all_flies[arena.identity] = row

                self.found_flies += 1

                # Draw the fly
                frame_color = fly.draw(frame_color, self.frame_count)

                # Add the fly to the arena
                arena.fly = fly

            ########################################
            ## End for loop over all putative arenas

            self.frame_color = frame_color


            # Update GUI graphics
            for i, arena in enumerate(sorted_arenas_br_to_tl_horizontally):
                arena_crop = arena.crop(frame_color)
                self.stacked_arenas[i] = imutils.resize(arena_crop, width=arena_crop.shape[1]*3)

            # TODO: Make into utils function
            #self.control_thread.stacked_arenas = self.stack_arenas(arenas_dict)

            # Print warning if number of flies in new frame is not the same as in previous
            if self.old_found != self.found_flies:
                logger.debug("Found %d flies in frame %d", self.found_flies, self.frame_count)
            self.old_found = self.found_flies
            self.found_flies = 0

            self.frame_count += 1


            return True

        #########################
        ## End if
        ##

        else:
            #self.save_prompt()
            return None


    @property
    def frame_count(self):
        return self._frame_count

    @frame_count.setter
    def frame_count(self, value):
        self._frame_count = value
        if self._frame_count % 100 == 0:
            logger.info("Frame #%s", (str(self._frame_count).zfill(5)))

    @property
    def ready(self):
        return self._ready_event.is_set()

    @ready.setter
    def ready(self, value):
        r"""
        Set me from an owning class so the behavior of track changes.
        """
        if value is True:
            self._ready_event.set()


    @property
    def framerate(self):
        return self._framerate

    @framerate.getter
    def framerate(self):

        # if not even started yet!
        if not self.running:
            framerate = 0

        # elif live feed
        elif self.video_path is None:
            framerate = self.camera.framerate

        # if video
        elif self.running and not self.stopped:
            before = self.frame_count
            self._end_event.wait(self._delta_t)
            after = self.frame_count
            framerate = (after - before) / self._delta_t


        self._framerate = framerate
        return self._framerate


    @property
    def info(self):
        return self._info

    @info.getter
    def info(self):
        logging.debug('Requesting SimpleTracker.info')
        self._info.update({
            "more_info": "more_info"
        })
        return self._info

    def run(self):
        """
        Start tracking in a while loop
        Munge processed frames and masks
        When while is finished, run self.close()
        Signal the exit event has been set if not yet
        This is the target function of the intermediate thread
        and it runs self.track an indefinite amount of time.
        """

        super().run()

        while not self.stopped:

            try:
                if self.track() is not True:
                    break

            except (LessArenasThanTargetError, CouldNotReadFrameError):
                self.fraction_area *= 0.99

        logger.debug("SimpleTracker.run detected stopped signal")

    def stop(self):
        """
        Release thre camera
        Save the cached data
        """
        super().stop()
        self.camera.stop()
        # TODO Move this to the control thread
        # self.control_thread.saver.store_and_clear()



if __name__ == "__main__":
    import argparse

    ap = argparse.ArgumentParser()
    ap.add_argument("-v", "--video")
    ap.add_argument("-c", "--camera_name", type=str, default="opencv", help="camera source")
    ap.add_argument("-g", "--gui", type=str, help="tkinter/opencv")
    ARGS = vars(ap.parse_args())

    # TODO Implement a main function for testing or independent running



import re
import datetime
import time
import logging

class DateRangeError(Exception):
    pass

class Scheduler(object):
    def __init__(self, in_str):
        """
        Class to express time constrains.
        It parses a formated string to define a list of allowed time range.
        Then it can be used to assess if a date and time is within a valid range.
        This is useful to control stimulators and other utilities.

        :param in_str: A formatted string. Format described `here <https://github.com/gilestrolab/ethoscope/blob/master/user_manual/schedulers.md>`_
        :type in_str: str
        """
        date_range_str = in_str.split(",")
        self._date_ranges = []
        for drs in  date_range_str:
            dr = self._parse_date_range(drs)

            self._date_ranges.append(dr)
        self._check_date_ranges(self._date_ranges)

    def _check_date_ranges(self, ranges):
        all_dates = []
        for start,end in ranges:
            all_dates.append(start)
            all_dates.append(end)

        for i  in  range(0, len(all_dates)-1):
            if (all_dates[i+1] - all_dates[i]) <= 0:
                raise DateRangeError("Some date ranges overlap")
        pass

    def check_time_range(self, t = None):
        """
        Check whether a unix timestamp is within the allowed range.
        :param t: the time to test. When ``None``, the system time is used
        :type t: float
        :return: ``True`` if the time was in range, ``False`` otherwise
        :rtype: bool
        """
        if t is None:
            t= time.time()
        return self._in_range(t)

    def _in_range(self, t):
        for r in self._date_ranges:
            if r[1] > t > r[0]:
                return True
        return False

    def _parse_date_range(self, str):
        self._start_date = 0
        self._stop_date = float('inf')
        dates = re.split("\s*>\s*", str)

        if len(dates) > 2:
            raise DateRangeError(" found several '>' symbol. Only one is allowed")
        date_strs = []
        for d in dates:
            date_strs.append(self._parse_date(d))

        if len(date_strs) == 1:
            # start_date
            if date_strs[0] is None:
                out =  (0,float("inf"))
            else:
                out = (date_strs[0], float("inf"))

        elif len(date_strs) == 2:
            d1, d2 = date_strs
            if d1 is None:
                if d2 is None:
                    raise DateRangeError("Data range cannot inclue two None dates")
                out =  (0, d2)
            elif d2 is None:
                out =  (d1, float("inf"))
            else:
                out =  (d1, d2)
        else:
            raise Exception("Unexpected date string")
        if out[0] >= out[1]:
            raise DateRangeError("Error in date %s, the end date appears to be in the past" % str)
        return out
        
    def _parse_date(self, str):
        pattern = re.compile("^\s*(?P<date>[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2})\s*$")
        if re.match("^\s*$", str):
            return None
        if not re.match(pattern, str):
            raise DateRangeError("%s not match the expected pattern" % str)
        datestr = re.match(pattern, str).groupdict()["date"]
        return time.mktime(datetime.datetime.strptime(datestr,'%Y-%m-%d %H:%M:%S').timetuple())



import logging
import warnings
import time
import traceback

from pypylon import pylon
import cv2

from idoc.server.io.cameras import AdaptorCamera

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



class PylonCamera(AdaptorCamera):
    r"""
    Drive a Basler camera using pypylon.
    """

    def __init__(self, *args, timeout=5000, video_path=None, **kwargs):

        # TODO Put this stuff in the config
        self._max_failed_count = 10000
        self._timeout = timeout
        self._video_path = video_path
        super().__init__(*args, **kwargs)

    def is_last_frame(self):
        return False

    def _time_stamp(self):
        now = time.time()
        return now - self._start_time

    def is_opened(self):
        return self.camera.IsOpen()

    def _next_image(self):

        failed_count = 0
        # this while loop is designed to try several times,
        # not to be run continuosly
        # i.e. in normal conditions, it should be broken every time
        # the continously running loop is implemented BaseCamera.__iter__

        while self.camera.IsGrabbing():

            grab = self._grab()
            if grab.GrabSucceeded():
                # logger.debug("grab succeeded")
                img = grab.Array
                img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
                self._validate(img)
                if failed_count != 0:
                    logger.warning(f"Pylon has recovered. Frame retrieval continues")
                    failed_count=0
                return img
            else:
                logger.warning(f"Pylon could not fetch next frame. Trial no {failed_count}/{self._max_failed_count}")
                failed_count += 1
                if failed_count == self._max_failed_count:
                    template = "Tried reading next frame %d times and none worked. Exiting."
                    message = template % self._max_failed_count
                    logger.warning(message)
                    self.close()
                    return None

    def open(self):

        try:
            super().open()
            logger.debug("Initializing camera")
            self.camera = pylon.InstantCamera(pylon.TlFactory.GetInstance().CreateFirstDevice())
            logger.debug("Opening camera")
            self.camera.Open()
            # Print the model name of the camera.
            logger.info("Using device %s", self.camera.GetDeviceInfo().GetModelName())
            # self.camera.StartGrabbingMax(100) # if we want to limit the number of frames
            self.camera.MaxNumBuffer = 5
            self.camera.StartGrabbing(pylon.GrabStrategy_LatestImageOnly)
            grab_result = self.camera.RetrieveResult(self._timeout, pylon.TimeoutHandling_ThrowException)
            image = grab_result.Array
            logger.info("Pylon camera loaded successfully")
            logger.info("Resolution of incoming frames: %dx%d", image.shape[1], image.shape[0])

        except Exception as error:
            logger.error(error)
            logger.error(traceback.print_exc())
            self.reset()


    # called by BaseCamera.__exit__()
    def close(self):
        super().close()
        logger.debug("Closing camera")
        self.camera.Close()


    def restart(self):
        r"""
        Attempt to restart a Basler camera
        """
        NotImplementedError


    def _grab(self):
        try:
            grab_result = self.camera.RetrieveResult(self._timeout, pylon.TimeoutHandling_ThrowException)
            return grab_result

        # TODO Use right exception
        except Exception as error:  # pylint: disable=broad-except
            logger.warning("Error in _grab method")
            logger.warning(traceback.print_exc())
            logger.warning(error)
            return


    @property
    def resolution(self):
        r"""
        Convenience function to return resolution of camera.
        Resolution = (number_horizontal_pixels, number_vertical_pixels)
        """
        self._settings["resolution"] = (self.camera.Width.GetValue(), self.camera.Height.GetValue())
        return self._settings["resolution"]



    @property
    def shape(self):
        r"""
        Convenience function to return shape of camera
        Shape = (number_vertical_pixels, number_horizontal_pixels, number_channels)
        """
        # TODO Return number of channels!
        return (self.camera.Height.GetValue(), self.camera.Width.GetValue(), 1)

    @property
    def framerate(self):
        try:
            return self.camera.AcquisitionFrameRateAbs.GetValue()
        except Exception:
            try:
                return self.camera.AcquisitionFrameRate.GetValue()
            except Exception:
                warnings.warn(
                    f"{self} camera doesn't have AcquisitionFrameRate or AcquisitionFrameRateAbs" \
                    "Please update the {self.__class__}.framerate property. I will assume 30"
                )
                return 30
        



r"""
Control each pin separately and concurrently by providing it with
start and end times and the value it should exhibit when on
These classes are instantiated by a Programmer instance
taking the user input and the information from a Controller instance
"""
# Standard library imports
import logging
from threading import Event, BrokenBarrierError
import time

import numpy as np

from idoc.debug import IDOCException
from idoc.server.core.base import Base, StatusThread

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
#pylint: disable=W0223

class BaseControllerThread(Base):
    r"""
    Abstract class for parallel execution of pins in a Dummy or Pyfirmata board
    This implementation requires each event is driven by a separate thread.
    """

    _modes = {"output": "o", "pwm": "p"}
    pin_state = None

    def __init__(
            self, i: int, hardware: str, pin, pin_state: dict, mode: str, start, end, value: float, *args,
            sampling_rate=10.0, result_writer=None, use_wall_clock=False, **kwargs
        ):
        """

        :param i: A unique identifier for each instance
        :type i: int

        :param hardware:
        Identify hardware component a thread instance will be in charge of
        Not used in the code other than for logging. The variable that tells an instance which pin
        to control is pin_number.
        :type hardware: ``str`

        :param pin_state:
        A dictionary storing pairs of pin numbers and their corresponding state
        i.e. turned on or off.
        It is updated live and shared by all the threads in run in the same Controller class
        This way, it can be used to query the state of the whole Controller.

        :param mode:
        A valid mode for the get_pin function.
        Eiher i (input), o (output) or p (pwm).

        :param start:
        Time in seconds when the thread should turn on a pin.
        :type start: ``float``

        :param end:
        Time in seconds when the thread should turn off a pin.
        :type end: ``float``

        :param value:
        Value to which a pin is set to. If equal to 0.0 or 1.0,
        it will be coerced to 0 or 1 and the pin will open in digital mode.
        If it is a float between 0 and 1,
        it is not coerced and the pin is opened in pwm mode with that intensity
        (max intensity is 1).

        :type value: ``float``

        :param sampling_rate:
        Frequency in Hertz (s-1) with which the thread will check
        if it should move on in the target function. Default 10.
        :type sampling_rate: ``float``

        :param args:
        Extra positional arguments for the Thread class.
        :type args: ``tuple``

        :param kwargs:
        Extra keyword arguments for the Thread class.
        :type kwargs: ``dictionary``
        """
        super().__init__(*args, **kwargs)

        self.index = i
        self._hardware = hardware
        self._pin = pin
        self.pin_state = pin_state
        self._barriers = {}
        self._result_writer = result_writer
        self._mode = mode
        self._start = start
        self._end = end
        self._value = value
        self.sampling_rate = sampling_rate
        self._last_t = 0
        self._use_wall_clock = use_wall_clock



    @property
    def start_seconds(self):
        return self._start

    @property
    def end_seconds(self):
        return self._end

    @property
    def hardware(self):
        return self._hardware

    @property
    def pin(self):
        return self._pin

    @property
    def name(self):
        r"""
        Unique thread identifier
        with format HARDWARE-i
        """
        return "%s-%s" % (self._hardware, self.index)

    @property
    def duration(self):
        r"""
        Seconds since the thread will first turn on the pin
        until it turns it off for the last time.
        """
        return self.end_seconds - self.start_seconds

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):

        try:
            assert 0 <= value <= 1
        except AssertionError:
            raise IDOCException('Passed value not valid: %s' % str(value))

        self._value = value

    @property
    def mode(self):
        return self._mode

    def add_barrier(self, barrier, name):
        r"""
        Set barrier as a blocking element during execution of the thread
        in the name position
        name = start -> before turning the pin on for the first time
        name = end -> after turning the pin off for the last time
        (right before self.run() is over).
        """
        self._barriers[name] = barrier

    def is_wave(self):
        r"""
        If True, the thread implements a duty cycle
        during the execution of self.turn_on()
        This is currently determined by the name of the class
        starting with Wave
        """
        value = self.__class__.__name__[:4] == "Wave"
        return value

    def turn_on(self, value=None):
        # TODO This will set the value of the pin in the convenience monitor
        # 'pin_state' to self.value for the whole period from start to end
        # even if the pin is controlled with a wave
        # Shall I change this so the monitor displays the wave?

        logger.info("%s: Turning %s on (%s)", self.last_t, self._hardware, value or self.value)
        self._notify(value or self.value)

        return self._turn_on(value or self.value)

    def turn_off(self):

        logger.debug(f"{round(self.last_t)}: Turning {self._hardware}  off (0)")
        self._notify(0)
        return self._turn_off()

    def _turn_on(self, value=None):
        # implemented in a subclass
        raise NotImplementedError

    def _turn_off(self, value=None):
        # implemented in a subclass
        raise NotImplementedError


    def _notify(self, value):
        self.pin_state[self._hardware] = value
        logger.info("%s notifying %s", self.hardware, str(value))

        self._result_writer.process_row(
            self.pin_state,
            table_name="CONTROLLER_EVENTS"
        )

    @property
    def last_t(self):
        return self._last_t

    @last_t.setter
    def last_t(self, last_t):
        self._last_t = last_t

    @last_t.getter
    def last_t(self):
        if self._use_wall_clock:
            return self.elapsed_seconds
        else:
            # Use the clock of the recognizer
            # Very useful when analyzing a video offline
            # so the video and the controller get aligned
            # Need to use --no-wall-clock-time
            return self._last_t

    def run(self):


        super().run()

        logger.debug(
            "%s (class %s) is starting to run",
            self.name, self.__class__.__name__
        )

        logger.debug(
            f"{self} will start @ {self.start_seconds} seconds"
        )

        while self.last_t < self.start_seconds:
            time.sleep(1 / self.sampling_rate)

            if self.stopped:
                logger.info('%s: %s is stopping before reaching the start barrier', self.last_t, self.name)
                self.turn_off()
                return


        # make sure all contemporaneous threads start as close to each other as possible
        # this call locks the thread until the contemporaneous threads reach the same state
        logger.debug(
            "%s: %s (class %s) reached the start barrier",
            self.last_t, self.name, self.__class__.__name__
        )

        try:
            self._barriers["start"].wait()
        except (KeyError, BrokenBarrierError):
            logger.info('%s: %s is stopping before turning on', self.last_t, self.name)
            return None

        self.turn_on()

        # wait until you should turn off
        if np.isnan(self.end_seconds):
            self._end_event.wait()
        else:
            while self.last_t < self.end_seconds:
                time.sleep(1 / self.sampling_rate)
                if self.stopped:
                    logger.info('%s: %s is stopping before turning off', self.last_t, self.name)
                    self.turn_off()
                    return

        # turn off
        self.turn_off()

        # signal to other threads ending simultaneously
        # and those that should start right afterwards
        logger.debug(
            "%s: %s (class %s) reached the end barrier",
            self.last_t, self.name, self.__class__.__name__
        )

        try:
            self._barriers["end"].wait()
        except (KeyError, BrokenBarrierError):
            logger.info('%s: The end barrier in %s was aborted', self.last_t, self.name)
            return

    def abort(self):
        r"""
        Stop execution by setting the end_event and aborting barriers.
        This invalidates the barriers, which means the threads won't block
        when they reach them.
        """
        for barrier in self._barriers.values():
            barrier.abort()

    def stop(self):
        logger.debug("Stopping thread")
        super().stop()
        logger.debug("Aborting barriers")
        self.abort()
        logger.debug("Turning off")
        self.turn_off()

class WaveBaseControllerThread(BaseControllerThread):
    r"""
    A Thread implementing a duty cycle pin as opposed to a
    continuously on pin
    The pin "blinks" when it is signalled to run
    Instead of being set to on for the whole duration of the period
    until turn_off is called,
    it turns on for on seconds and off for off seconds.
    """

    def __init__(self, seconds_on, seconds_off, *args, **kwargs):
        r"""
        :param seconds_on: Time in seconds the pin should stay on in each cycle.
        :type seconds_on: ``float``

        :param seconds_off: Time in seconds the pin should stay off in each cycle.
        :type seconds_off: ``float``
        """
        self._seconds_on = seconds_on
        self._seconds_off = seconds_off
        self._shore = Event()
        self._thread_wave = StatusThread(target=self.wave)
        super().__init__(*args, **kwargs)


    @property
    def shore(self):
        return self._shore.is_set()

    @shore.setter
    def shore(self, value):
        if not self._shore.is_set():
            if value is True:
                self._shore.set()

    def stop(self):
        self.shore = True
        super().stop()

    def reset(self):
        super().reset()
        self._shore.clear()


    def get_thread_wave(self, value=None):
        if self._thread_wave.stopped:
            self.reset()
            self._thread_wave = StatusThread(target=self.wave, kwargs={"value": value})

        return self._thread_wave


    def wave(self, value=None):
        """
        Turn on the pin, wait on seconds, turn off the pin, wait off seconds and start over
        as long as the _shore event is not set
        """


        while not self._shore.is_set():
            self._turn_on(value or self.value)
            self._shore.wait(self._seconds_on)
            self._turn_off()
            self._shore.wait(self._seconds_off)


    def turn_on(self, value=None):

        if value is not None:
            self.value = value
            self.get_thread_wave(value or self.value)

        if not self._thread_wave.running:
            logger.info("%s: Turning %s on (%s)", self.last_t, self._hardware, value or self.value)

            self._notify(value or self.value)
            self._thread_wave.start()

    def turn_off(self):

        logger.info("%s: Turning %s off (0)", self.last_t, self._hardware)
        self._thread_wave.stop()
        self._notify(0)
        self._shore.set()


class ArduinoMixin():

    r"""
    Mixin encapsulating all the Arduino/Pyfirmata functionality
    Add it to classes BaseControllerThread or WaveBaseControllerThread to
    'teach' them how to interact with Arduino boards using pyfirmata.

    * Selection of pins
    * Turn on and off a pin
    """

    # supplied in the classes where ArduinoMixin is mixed in
    _value = 1
    mode = None

    def _turn_on(self, value=None):
        self.pin.write(value or self.value)

    def _turn_off(self):
        self.pin.write(0)

class DummyMixin():
    r"""
    Mixin encapsulating dummy functionality
    Add it to classes BaseControllerThread or WaveBaseControllerThread to
    show them how to just log to the terminal and do nothing
    for dev/debugging purposes
    """

    def _turn_on(self, value=None):
        if not self.stopped:
            self.pin.write(value or self.value)
            # logger.info(
            #     "%s (class %s) is setting %s to %.8f @ %.4f",
            #     self.name, self.__class__.__name__, self._hardware, value or self._value, self.elapsed_seconds or 0
            # )


    def _turn_off(self):
        self.pin.write(0)

        if not self.stopped:
            pass
            # logger.info(
            #     "%s (class %s) is setting %s to %.8f @ %.4f",
            #     self.name, self.__class__.__name__, self._hardware, 0, self.elapsed_seconds or 0
            # )

class DefaultDummyThread(DummyMixin, BaseControllerThread):
    r"""
    A dummy thread that just logs without actually activating any hardware
    Useful for debugging
    """

class WaveDummyThread(DummyMixin, WaveBaseControllerThread):
    r"""
    A dummy thread that just logs without actually activating any hardware
    Useful for debugging
    """

class DefaultArduinoThread(ArduinoMixin, BaseControllerThread):
    r"""
    A thread using an Arduino board without wave
    """

class WaveArduinoThread(ArduinoMixin, WaveBaseControllerThread):
    r"""
    A thread using an Arduino board with wave
    """



r"""
Boards that a Controller class can drive with the program generated
in a Programmer class.
For now, only a dummy board that just logs to the screen and adaptor classes
from pyfirmata are available.
"""
import logging


from pyfirmata import Arduino, ArduinoMega

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
# pylint: disable=too-few-public-methods

class PIN():

    def __init__(self, signal, number, mode):
        r"""
        """
        self._signal = signal
        self._number = number
        self._mode = mode

    def write(self, value):
        logger.debug(f"Setting pin {self._number} -> {value} (mode {self._mode})")

class ArduinoDummy:
    """
    A dummy class to be passed to a Controller class for debugging purposes
    """

    def __init__(self, devport):
        self._devport = devport
        logger.info("Initialized dummy Arduino on port %s", self._devport)


    def get_pin(self, string):

        signal = string.split(":")[0]
        number = int(string.split(":")[1])
        mode = string.split(":")[2]

        pin = PIN(signal, number, mode)
        return pin

    def exit(self):
        pass

# TODO Remove these classes and just use pure pyfirmata
class ArduinoBoard(Arduino):
    pass

class ArduinoMegaBoard(ArduinoMega):
    pass



"""
Track flies, record their position and draw the incoming frames
"""
import datetime
import logging
import os.path
import time
import traceback


from idoc.server.core.tracking_unit import TrackingUnit
from idoc.server.core.variables import FrameCountVariable
from idoc.server.core.base import Base
from idoc.server.roi_builders.roi_builders import DefaultROIBuilder
from idoc.debug import IDOCException, EthoscopeException
from idoc.helpers import iso_format, hours_minutes_seconds

from idoc.configuration import IDOCConfiguration
config = IDOCConfiguration()

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.INFO)


class Recognizer(Base):

    valid_actions = ["start", "stop", "prepare", "reset"]

    def __init__(
            self, tracker_class, camera_class,
            roi_builder, result_writer, drawer_class, drawer_kwargs,
            rois=None, stimulators=None,
            user_data=None
        ):
        """
        Class to orchestrate the tracking of multiple objects.
        It performs, in order, the following actions:

         * Requesting raw frames (delegated to :class:`~ethoscope.hardware.input.cameras.BaseCamera`)
         * Cutting frame portions according to the ROI layout (delegated to :class:`~ethoscope.core.tracking_unit.TrackingUnit`).
         * Detecting animals and computing their positions and other variables (delegated to :class:`~ethoscope.trackers.trackers.BaseTracker`).
         * Using computed variables to interact physically (i.e. feed-back) with the animals (delegated to :class:`~ethoscope.stimulators.stimulators.BaseStimulator`).
         * Drawing results on a frame, optionally saving video (delegated to :class:`~ethoscope.drawers.drawers.BaseDrawer`).
         * Saving the result of tracking in a database (delegated to :class:`~ethoscope.utils.io.ResultWriter`).

        :param camera: a camera object responsible of acquiring frames and associated time stamps
        :type camera: :class:`~ethoscope.hardware.input.cameras.BaseCamera`
        :param tracker_class: The algorithm that will be used for tracking. It must inherit from :class:`~ethoscope.trackers.trackers.BaseTracker`
        :type tracker_class: class
        :param rois: A list of region of interest.
        :type rois: list(:class:`~ethoscope.core.roi.ROI`)
        :param stimulators: The class that will be used to analyse the position of the object and interact with the system/hardware.
        :type stimulators: list(:class:`~ethoscope.stimulators.stimulators.BaseInteractor`)
        :type stimulators: list(:class:`~ethoscope.stimulators.stimulators.BaseInteractor`)
        :param args: additional arguments passed to the abstract classes
        :param kwargs: additional keyword arguments passed to the abstract classes
        """

        super().__init__()
        self._user_data = user_data

        self._last_frame_idx = 0
        self._force_stop = False
        self._last_positions = {}
        self._last_time_stamp = 0
        self._last_tick = 0
        self._period = 1
        self._unit_trackers = []
        self._rois = rois
        self._frame_buffer = None
        self.rois = []

        self._add_submodule("result_writer", result_writer)
        self._submodules["camera"] = None
        self._submodules["drawer"] = None
        self._drawer_class = drawer_class
        self._drawer_kwargs = drawer_kwargs

        self._tracker_class = tracker_class
        self._camera_class = camera_class
        self._roi_builder = roi_builder
        self._stimulators = stimulators

        self._last_frames = []
        self._last_times = []
        self._time_running = False
        self._time_offset = 0
        self._frame_offset = 0



    @property
    def last_t(self):
        return self.last_time_stamp

    @property
    def info(self):
        self._info = {
            "status": self.status,
            "last_drawn_path": getattr(self, "drawer.last_drawn_path", None),
            "last_annot_path": getattr(self, "drawer.last_annot_path", None),
            "camera": self.camera.__class__.__name__
        }
        return self._info

    def load_camera(self):
        r"""
        Instantiate a camera object
        which can be used to sample frames
        and recognize flies.
        This way the camera is actually not initialized
        until this method is called in the recognizer.prepare method
        i.e. not at Recognizer.__init__.
        """
        camera_args = config.content['io']['camera']['args']
        camera_kwargs = config.content['io']['camera']['kwargs']

        if self._submodules["camera"] is None or self._submodules["camera"].is_opened():
            logger.debug("Loading camera")
            user_data = self._user_data.copy()

            # Pass settings the user may have passed upon starting the program
            # kwargs to be passed on initialization of camera class
            camera_kwargs["video_path"] = config.content['io']['camera']['kwargs']['video_path']
            camera_kwargs["max_duration"] = config.content['experiment']['max_duration']
            camera_kwargs["use_wall_clock"] = config.content['io']['camera']['kwargs']['use_wall_clock']


            self._submodules["camera"] = self._camera_class(
                *camera_args,
                **camera_kwargs
            )
        
            # Pass settings the user may have passed upon starting the program
            # kwargs that can be passed afterwards

            if user_data is not None:
                for setting in user_data:
                    if setting in self._submodules["camera"].settings:
                        if user_data[setting] is not None:
                            # logger.debug("Assigning ", setting, "to camera")
                            self._submodules["camera"].settings[setting] = user_data[setting]
                        else:
                            pass
                            # logger.debug("Not assigning ", setting, "to camera. Input value is None")

                    else:
                        pass
                        # logger.debug("Not assigning ", setting, "to camera. Setting does not belong")

            self._settings['camera'] = self.camera.settings

    def _build(self):
        """
        Build the ROIs used in the experiment
        If it fails, emit a warning and use the whole frame as the ROI
        """
        try:
            rois = self._roi_builder.build(self.camera)

        except (IDOCException, EthoscopeException) as error: # TODO remove ethoscopexception dependency
            logger.warning(traceback.print_exc())
            logger.warning(error)
            roi_builder_class = DefaultROIBuilder
            roi_builder = roi_builder_class()
            rois = roi_builder.build(self.camera)
            logger.debug("Default ROI loaded")

        return rois


    def _load(self, *args, rois=None, **kwargs):
        r"""
        Populate the unit_trackers slot which is iteratively gone through
        on every loop in run, to track every ROI independently
        """

        if rois is None:
            rois = self._rois
            if rois is None:
                raise NotImplementedError("rois must exist (cannot be None)")

        if self._stimulators is None:

            self._unit_trackers = [TrackingUnit(self._tracker_class, r, None, *args, **kwargs) for r in rois]

        elif len(self._stimulators) == len(rois):
            self._unit_trackers = [TrackingUnit(self._tracker_class, r, inter, *args, **kwargs) for r, inter in zip(rois, self._stimulators)]
        else:
            raise ValueError("You should have one interactor per ROI")

    def set_drawer(self):
        
        self._drawer_kwargs["framerate"] = self._submodules["camera"].framerate
        drawer = self._drawer_class(
            **self._drawer_kwargs
        )
        self._add_submodule("drawer", drawer)

    def prepare(self, *args, **kwargs):
        """
        Open the camera, build the ROIs and bind them to the trackers
        """
        self.load_camera()
        self.set_drawer()
        self.rois = self._build()
        self._load(*args, rois=self.rois, **kwargs)
        self.ready = True

    @property
    def camera(self):
        return self._submodules["camera"]

    @property
    def drawer(self):
        return self._submodules["drawer"]

    @property
    def result_writer(self):
        return self._submodules["result_writer"]

    @property
    def last_positions(self):
        """
        :return: The last positions (and other recorded variables) of all detected animals
        :rtype: dict
        """
        return self._last_positions

    @property
    def last_time_stamp(self):
        """
        :return: The time, in seconds, since recognizer started running. It will be 0 if the monitor is not running yet.
        :rtype: float
        """
        self._last_time_stamp
        return self._last_time_stamp

    @last_time_stamp.setter
    def last_time_stamp(self, last_time_stamp):

        last_time_stamp_s = last_time_stamp / 1e3

        if len(self._last_times) == 10:
            self._last_times.pop(0)

        self._last_times.append(last_time_stamp_s)

        if self._time_running:
            self._last_time_stamp = last_time_stamp_s - self._time_offset
        else:
            self._time_offset = last_time_stamp_s


    @property
    def last_frame_idx(self):
        """
        :return: The number of the last acquired frame.
        :rtype: int
        """
        return self._last_frame_idx

    @last_frame_idx.setter
    def last_frame_idx(self, last_frame_idx):
        """
        :return: The number of the last acquired frame.
        :rtype: int
        """
        if len(self._last_frames) == 10:
            self._last_frames.pop(0)

        self._last_frames.append(last_frame_idx)

        if self._time_running:
            self._last_frame_idx = last_frame_idx - self._frame_offset
        else:
            self._frame_offset = last_frame_idx

    @property
    def fps(self):

        try:
            end_f, start_f = (self._last_frames[::-1][0], self._last_frames[0])
            end_t, start_t = (self._last_times[::-1][0], self._last_times[0])
        except IndexError:
            return 0

        try:
            framerate = (end_f - start_f) / (end_t - start_t)
        except ZeroDivisionError:
            framerate = 0

        return framerate

    def run(self):
        """
        Runs the monitor indefinitely.
        """
        

        if self.ready is not True:
            return

        super().run()

        try:
            logger.info("Recognizer running")

            for items in enumerate(self.camera):

                i, (t_ms, frame) = items

                if self.result_writer is not None:
                    self.result_writer.last_t = self.last_t

                # logger.warning('Reading  frame at %d', t_ms)

                if self.stopped:
                    logger.info("Recognizer object stopped from external request")
                    break

                self.last_frame_idx = i
                self.last_time_stamp = t_ms
                self._frame_buffer = frame

                # if quality_controller is not None:
                #     qc = quality_controller.qc(frame)
                #     quality_controller.write(t, qc)

                for _, track_u in enumerate(self._unit_trackers):
                    data_rows = track_u.track(t_ms, frame)
                    if len(data_rows) == 0:
                        self._last_positions[track_u.roi.idx] = []
                        continue

                    abs_pos = track_u.get_last_positions(absolute=True)

                    self._last_positions[track_u.roi.idx] = abs_pos

                    if self.result_writer is not None:
                        frame_count = FrameCountVariable(i)
                        data_rows[0].append(frame_count)
                        self.result_writer.write(
                            # pass roi to tell the result writer the fly id
                            track_u.roi,
                            # actual tracking data
                            data_rows
                        )

                metadata = {
                    "time": iso_format(hours_minutes_seconds(self.last_t)),
                    "frame": self._last_frame_idx,
                    "framerate": round(self.fps)
                }

                if self.drawer is not None:
                    annot = self.drawer.draw(frame, tracking_units=self._unit_trackers, positions=self._last_positions, metadata=metadata)
                    tick = int(round((t_ms / 1000.0) / self._period))
                    if self._time_running:
                        video_root = os.path.join(
                            self.result_writer.result_dir,
                            "%s.avi" % self.result_writer.run_id_long
                        )
                        self.drawer.write_videos(video_root=video_root)

                    if tick > self._last_tick:
                        # logger.debug("Writing frame %d", i)
                        self.drawer.write_frames()
                        self.result_writer.write_frames(frame, self.drawer.frame_count)

                    self._last_tick = tick

        except Exception as error:
            logger.warning("Recognizer closing with an exception")
            logger.warning(traceback.format_exc())
            raise error

        finally:
            logger.info("Recognizer closing")
            self.stop()

    def stop(self):
        super().stop()

        if self.camera is not None:
            self.camera.close()
        if self.drawer is not None:
            self.drawer.close()

    def reset(self):
        super().reset()
        # TODO Even after redoing the instance by calling __init__
        # I still get
        # RuntimeError: threads can only be started once

        # self.__init__(
        #     self._tracker_class, self._camera_class,
        #     self._submodules['roi_builder'], self._submodules['result_writer'], self._submodules['drawer'], *self._args,
        #     rois=self._rois, stimulators=self._stimulators, video_path=self._video_path, **self._kwargs
        # )

if __name__ == "__main__":

    import argparse

    from idoc.server.trackers.adaptive_bg_tracker import AdaptiveBGModel
    from idoc.server.io.pylon_camera import PylonCamera
    from idoc.server.io.opencv_camera import OpenCVCamera
    from idoc.server.io.result_writer import CSVResultWriter
    from idoc.server.roi_builders.target_roi_builder import IDOCROIBuilder
    from idoc.server.drawers.drawers import DefaultDrawer

    def main():

        parser = argparse.ArgumentParser()
        parser.add_argument("-c", "--camera", type=str, choices=["PylonCamera", "OpenCVCamera"], default="OpenCVCamera")
        ARGS = vars(parser.parse_args())

        camera_class_mame = ARGS["camera"]

        for camera_class in [PylonCamera, OpenCVCamera]:
            if camera_class_mame == camera_class.__name__:
                break

        roi_builder = IDOCROIBuilder()
        result_writer = CSVResultWriter(start_datetime="2000-01-01_00-00-00")
        drawer = DefaultDrawer(draw_frames=False)

        recognizer = Recognizer(
            AdaptiveBGModel, camera_class,
            roi_builder, result_writer, drawer
        )

        recognizer.prepare()

    main()



# https://stackoverflow.com/a/4761058
import sys
import unittest
from pathlib import Path
import logging
import os
import threading
import time
import signal
from tests import test_generic
from src.interface.main import Interface
sys.path.append('../src')
from definitions import ROOT_DIR

class TestLiveStream(unittest.TestCase, Interface):

    def __init__(self, *args, **kwargs):


        super().__init__(*args, **kwargs)
        Interface.__init__(self,
        #arduino=True,
        #mapping=Path(ROOT_DIR, "mappings/main.csv"), program=Path(ROOT_DIR, "programs/simplified_program5.csv"), port="/dev/ttyACM0",
        track=True,
        camera='opencv', gui="tkinter", video = None,
        reporting=False, config=Path(ROOT_DIR, "config.yml"), duration=120
        )
        self.prepare()

    def send_signal(self, pid):
        time.sleep(.5)
        os.kill(pid, signal.SIGINT)

    def launch_and_send_exit(self):

        pid = os.getpid()
        press_control_c_thread = threading.Thread(
            name='press_control_c',
            target=self.send_signal,
            args=([pid]),
        )
        
        press_control_c_thread.start()
        self.start()

    
    def test_exit_signal(self):
        self.launch_and_send_exit()
        self.exit.wait()
        # give time to stop once signal is received
        time.sleep(.2)

        #self.assertFalse(self.arduino_thread.is_alive())
        self.assertFalse(self.tracker_thread.is_alive())

        
    # def test_camera_exists(self):
    #     test_result = False:
        
    #     try:
    #         Tracker(self, self.canera, self.video)
    #         test_result = True
    #



#! /home/vibflysleep/anaconda3/envs/idoc/bin/python

# Standard library imports
import logging
import math
import os
import glob
import sys
import signal
import subprocess
from threading import Thread, Event
import tempfile
import time
import traceback
import urllib.request, urllib.error, urllib.parse
import warnings
import webbrowser

# Local application imports
import cv2
import numpy as np
import pandas as pd

from idoc.configuration import IDOCConfiguration
from idoc.helpers import get_machine_id
from idoc.client.utils.device_scanner import DeviceScanner
from idoc.debug import IDOCException

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

config = IDOCConfiguration()

class CliUI():
    """
    Control IDOC interactively with a command line interface
    sending HTTP requests to the server API on behalf of the user.
    """

    _last_img_annot = "/tmp/last_img_annot.png"

    def __init__(self, device_id):
        self._device_id = device_id
        self._ds = DeviceScanner(port=config.content["network"]["port"])
        self._ds.start()
        self._answer = 0
        # self._func = None
        self.running = False
        self.stopped = False

        self._menus = {
            "root": [
                ('LOAD PARADIGM', None, "paradigms"),
                ('CHANGE CONFIGURATION', None, "settings"),
                ('WARM UP', self.warm_up, "root"),
                ('CHECK', self.check, "root"),
                ('START', self.start_experiment, "root"),
                # ('ANALYZE', self.analyze, "root"),
                ('STOP', self.stop, "root"),
                ('CLEAR', self.clear, "root"),
                ('PROMPT', self.prompt, "root"),
                ('EXPORT', self.export, "root"),
                ('RESTART', self.restart, "root")
                # ('RESET', self.reset, "root"),
                # ('QUIT', self.close, "root"),

            ],

            "settings" : [
                ['recognizer', 'camera', 'framerate'],
                ['recognizer', 'camera', 'exposure_time'],
                ['adaptation_time'],
                []
            ]
        }

        self._pick_dict = {"root": self._pick_menu, "settings": self._pick_setting, "paradigms": self._pick_paradigm}

        # give time for the device scanner to find the device
        time.sleep(2)
        self.initialise_device()


    def initialise_device(self):
        try:
           self._device = self._ds.get_device_by_id(self._device_id)
        except Exception as error:
            logger.error(error)
            time.sleep(1)
            self.initialise_device()


    @property
    def paradigms(self):
        try:
            print(self._device.info)
            paradigms = self._device.info["controller"]["paradigms"]
            paradigms = [e for e in paradigms.values()]
            paradigms.append(None)
            return paradigms

        except AttributeError:
            return [None]

    @staticmethod
    def represents_int(s):
        try:
            int(s)
            return True
        except ValueError:
            return False


    # Core functionality
    ## Methods needed to run the CliUI
    ## * _pick_menu: ask the user to perform an action
    ## * _pick_setting: ask the user to enter an option or some input
    ## * _apply: apply or execute the input provided by the user
    ## * run: iterate a new question and execution


    def _user_input(self, max_index, prompt="Enter number: "):
        answer = input(prompt)
        fail = False
        fail = not self.represents_int(answer)
        if not fail:
            answer = int(answer)
            fail = answer > max_index

        if fail:
            print('You entered answer: %s' % answer)
            print('This is not valid. Try again!')

            answer = self._user_input(prompt, max_index)

        return answer

    def _pick_menu(self):
        """
        Present the user with a list of menus
        """

        print("Please choose:")
        options = self._menus["root"]
        for idx, opt in enumerate(options):
            print("{}: {}".format(idx + 1, opt[0]))

        self._answer = self._user_input(len(options))
        return self._answer


    def _pick_setting(self):

        print("Please choose:")
        options = self._menus["settings"]
        for idx, opt in enumerate(options):
            print("{}: {}".format(idx + 1, '/'.join(opt)))

        self._answer = self._user_input(len(options))
        return self._answer

    def _pick_paradigm(self):
        print("Please choose:")
        options = self.paradigms
        for idx, opt in enumerate(options):
            print("{}: {}".format(idx + 1, opt))

        self._answer = self._user_input(len(options))
        return self._answer


    def _apply(self, menu_name, answer):
        """
        Apply the selected option and achieve a side-effect.
        """
        # answer is 0 when an error happpened in the previous pick operation
        if answer == 0:
            return "root"


        # FIXME Weird handling when we are loading a paradigm
        if menu_name == "paradigms":
            options = self.paradigms
        else:
            options = self._menus[menu_name]

        opt = options[answer - 1]

        # the last option is always reserved for a return/stop operation
        # that has no side effects (only effects within the core CLI)
        # handled by self.run
        if answer == len(options) and menu_name != "root":
            new_menu = "root"

        elif menu_name == "root":
            name, func, new_menu = opt
            if new_menu == "root":
                new_menu = func(answer)

            else:
                if new_menu == "settings":
                    # ignore opt and just get a new answer
                    answer = self._pick_setting()
                elif new_menu == "paradigms":
                    answer = self._pick_paradigm()

                new_menu = self._apply(new_menu, answer)
                # open the new menu

        else:

            if menu_name == "settings":
                self._change_setting(answer)
                answer = self._pick_setting()

            elif menu_name == "paradigms":
                self._load_paradigm(answer)
                answer = self._pick_paradigm()

            else:
                # TODO Warn the user
                pass

            new_menu = self._apply(menu_name, answer)

        return new_menu


    def run(self, menu_name="root"):
        """
        Everytime this method is called
        * the user picks an option from the current menu
        * the option is applied (we get a side-effect)
        * a new menu is chosen automatically
        """

        self.running = True

        answer = self._pick_menu()

        while not self.stopped:
            try:
                new_menu = self._apply(menu_name, answer)
                func = self._pick_dict[new_menu]
                answer = func()

            except Exception as error:
                warnings.warn('I do not understand %s' % self._answer)
                print(error)
                print(traceback.print_exc())
                new_menu = "root"
                answer = 0



    # Side effect functionality
    ## Methods needed to apply the wishes of the user

    def _load_paradigm(self, answer):
        """
        Receive an answer and post the corresponding paradigm_path to the device
        """

        paradigm_path = self.paradigms[answer - 1]
        data = {"paradigm_path": paradigm_path}
        self._device.post_paradigm(data)

        return "paradigms"

    def warm_up(self, answer):
        """
        Communicate with the device to warm_up.

        Start the recognizer and the always_on_hardware
        without saving data.
        Open a broser tab to view the flies
        """
        self._device.warm_up()
        webbrowser.open_new_tab("file://%s" % self._last_img_annot)

        return "root"


    def start_experiment(self, answer):
        """
        Communicate with the device to start an experiment.
        """
        self._device.start_experiment()

        return "root"

    def check(self, answer):
        """
        View the status of IDOC
        """
        warnings.warn("Not implemented")
        return "root"

    def analyze(self, answer):
        warnings.warn("Not implemented")
        return "root"

    @staticmethod
    def clear(self):
        for _ in range(30):
            print(" ")

        return "root"

    def prompt(self, answer):
        self._pick_menu()
        return "root"

    def quit(self, *args, **kwargs):
        self.stop("")


    def stop(self, answer):
        """
        "Ask the user if output should be saved
        and quit the program
        """
        # TODO Ask functionality
        self._device.stop_experiment()
        return "root"

    @staticmethod
    def get_run_id(experiment_folder):

        metadata_csv = glob.glob("%s/*_METADATA.csv" % experiment_folder)[0]
        metadata = pd.read_csv(metadata_csv)
        start_datetime = metadata.loc[metadata['field'] == 'date_time', "value"].values[0]
        machine_id = metadata.loc[metadata['field'] == 'machine_id', "value"].values[0]
        run_id = "%s_%s" % (start_datetime, machine_id)
        return run_id


    def _export_summary_csv(self, experiment_folder, run_id):
        output_csv = os.path.join(experiment_folder, '%s_SUMMARY.csv' % run_id)
        r_cmd = "idocr::export_summary('%s', '%s')" % (experiment_folder, output_csv)
        subprocess.call(["R", "-e", r_cmd])

    def _export_pi_data(self, experiment_folder, run_id):
        output_csv = os.path.join(experiment_folder, '%s_PI.csv' % run_id)
        r_cmd = "idocr::export_pi('%s', '%s')" % (experiment_folder, output_csv)
        subprocess.call(["R", "-e", r_cmd])

    def export(self, answer):
        """
        Save convenience csv files for analysis in external software
        """

        experiment_folder = input('Enter experiment_folder: ')
        if os.path.isdir(experiment_folder):
            run_id = self.get_run_id(experiment_folder)
            self._export_summary_csv(experiment_folder, run_id)
            self._export_pi_data(experiment_folder, run_id)


    def restart(self, answer):
        logger.info("Restarting idoc_server systemd service. Please give a few seconds...")
        subprocess.call(["systemctl", "restart", "idoc_server.service"])
        return "root"

    def reset(self, answer):

        self._device.reset_experiment()
        return "root"

    def close(self, answer):
        print("Quitting IDOC")
        self._ds.stop()
        self.stopped = True
        time.sleep(1)
        sys.exit(0)

    def _change_setting(self, answer):
        """
        Update the IDOC setting the user has selected with the answer
        Answer is an integer from 1 to length of self_menus['settings']
        """
        settings = self._menus["settings"].copy()

        selected_setting = settings[answer - 1]

        try:
            new_value = self.prompt_new_setting(selected_setting)
        except IDOCException:
            return

        print("Setting %s to %s" % ('/'.join(selected_setting), new_value))

        if len(selected_setting) == 1:
            setting_name = selected_setting[0]
            submodule = "control_thread"

        else:
            setting_name = selected_setting[::-1][0]
            submodule = selected_setting[:len(selected_setting) - 1]

        settings = {"settings": {setting_name: new_value}, "submodule": submodule}
        self._device.post_settings(settings)

        return "settings"

    def prompt_new_setting(self, setting_name):
        """
        Ask the user to enter a new value for setting setting_name
        """

        device_settings = self._device.info['settings']

        # if the setting name is a list, it means it is a submodule setting
        # and we need to dig in the device_settings dict until a match is found
        if isinstance(setting_name, str):
            pass

        elif isinstance(setting_name, list):
            submodule_setting = dict(device_settings)
            for level in setting_name:
                if level in submodule_setting:
                    submodule_setting = submodule_setting[level]
                    if not isinstance(submodule_setting, dict):
                        break
                else:
                    warnings.warn("%s is not configurable yet." % level)
                    warnings.warn("Hint: Did you start the module this setting controls?.")
                    raise IDOCException("Setting %s is not implemented" % level)

            device_settings = submodule_setting
            setting_name = level

        else:
            # TODO Ellaborate more info for the user here
            warnings.warn("Setting is not valid")
            warnings.warn(setting_name)
            return

        # Pick the value of the old setting and report it to the user
        print('%s is set to %s', setting_name, device_settings)
        # Ask the user to enter a new value
        value = float(input("Enter new %s: " % setting_name))

        # TODO Validation here

        return value


def main():
    device_id = get_machine_id()
    cli = CliUI(device_id=device_id)
    time.sleep(5)

    signals = ('TERM', 'HUP', 'INT')
    for sig in signals:
        signal.signal(getattr(signal, 'SIG' + sig), cli.quit)

    cli.run()

if __name__ == "__main__":
    main()



__author__ = 'quentin'

from collections import deque

from idoc.server.core.base import DescribedObject
from idoc.server.core.variables import IsInferredVariable
from idoc.debug import IDOCException

class NoPositionError(Exception):
    """
    Used to abort tracking. When it is raised within the ``_find_position`` method,
    data is inferred from previous position.
    """
    pass # pylint: disable=unnecessary-pass

class BaseTracker(DescribedObject):
    # data_point = None
    def __init__(self, roi, *args, data=None, **kwargs):
        """
        Template class for video trackers.
        A video tracker locate animal in a ROI.
        Derived class must implement the ``_find_position`` method.

        :param roi: The Region Of Interest the the tracker will use to locate the animal.
        :type roi: :class:`~ethoscope.rois.roi_builders.ROI`
        :param data: An optional data set. For instance, it can be used for pre-trained algorithms

        :return:
        """
        self._positions = deque()
        self._times = deque()
        self._data = data
        self._roi = roi
        self._last_non_inferred_time = 0
        self._last_time_point = 0
        self._max_history_length = 250 * 1000  # in milliseconds
        self._old_pos = 0.0 +0.0j

        super().__init__(*args, **kwargs)

        # self._max_history_length = 500   # in milliseconds
        # if self.data_point is None:
        #     raise NotImplementedError("Trackers must have a DataPoint object.")

    def track(self, t, img):
        """
        Locate the animal in a image, at a given time.

        :param t: time in ms
        :type t: int
        :param img: the whole frame with 3 channels
        :type img: :class:`~numpy.ndarray`
        :return: The position of the animal at time ``t``
        :rtype: :class:`~ethoscope.core.data_point.DataPoint`
        """

        if len(img.shape) != 3:
            raise IDOCException("tracker.track: input image is not 3D")

        sub_img, mask = self._roi.apply(img)
        self._last_time_point = t
        try:

            points = self._find_position(sub_img, mask, t)

            if not isinstance(points, list):
                raise Exception("tracking algorithms are expected to return a LIST of DataPoints")

            if len(points) == 0:
                return []

            # point = self.normalise_position(point)
            self._last_non_inferred_time = t

            for p in points:
                p.append(IsInferredVariable(False))

        except NoPositionError:
            if len(self._positions) == 0:
                return []
            else:

                points = self._infer_position(t)

                if len(points) == 0:
                    return []
                for p in points:
                    p.append(IsInferredVariable(True))

        self._positions.append(points)
        self._times.append(t)


        if len(self._times) > 2 and (self._times[-1] - self._times[0]) > self._max_history_length:
            self._positions.popleft()
            self._times.popleft()
        return points

    def _infer_position(self, t, max_time=30 * 1000):
        if len(self._times) == 0:
            return []
        if t - self._last_non_inferred_time > max_time:
            return []

        return self._positions[-1]

    @property
    def positions(self):
        """
        :return: The last few positions found by the tracker.\
            Positions are kept for a certain duration defined by the ``_max_history_length`` attribute.
        :rtype: :class:`~collection.deque`
        """
        return self._positions

    def xy_pos(self, i):
        return self._positions[i][0]

    @property
    def last_time_point(self):
        """
        :return: The last time point that the tracker used.\
            This is updated even when position is inferred/no animal is found
        :rtype: int
        """
        return self._last_time_point

    @property
    def times(self):
        """
        :return: The last few time points corresponding to :class:`~ethoscope.trackers.trackers.BaseTracker.positions`.
        :rtype: :class:`~collection.deque`
        """
        return self._times

    def _find_position(self, img, mask, t):
        raise NotImplementedError



import argparse
import datetime
import json
import logging
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import urllib.error

import bottle
import netifaces

from idoc.client.utils.device_scanner import DeviceScanner
from idoc.configuration import IDOCConfiguration
from idoc.decorators import warning_decorator, error_decorator
from idoc.helpers import get_server

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.INFO)

logging.basicConfig(
    format='%(relativeCreated)5d %(name)-15s %(levelname)-8s %(message)s'
)



app = bottle.Bottle()
STATIC_DIR = "../../static"
tmp_imgs_dir = tempfile.mkdtemp(prefix="idoc_images")


ap = argparse.ArgumentParser()
ap.add_argument("-p", "--port", type=int, default=8000)
ap.add_argument("-D", "--debug", action='store_true', dest="debug")
ARGS = vars(ap.parse_args())
CFG = IDOCConfiguration()

PORT = ARGS["port"]
DEBUG = ARGS["debug"]

logger.info("Running client on port %d", PORT)

ds = DeviceScanner()
ds.start()


@app.route('/')
def index():
    return bottle.static_file('index.html', root=STATIC_DIR)


@app.route('/static/<filepath:path>')
def server_static(filepath):
    return bottle.static_file(filepath, root=STATIC_DIR)

@app.get('/favicon.ico')
def get_favicon():
    return server_static(STATIC_DIR+'/img/favicon.ico')


@app.get('/devices')
@error_decorator
def devices():
    return ds.get_all_devices_info()


@app.get('/device/<id>')
def get_device_by_id(id):
    return {"id": ds.get_device_by_id(id).info["id"]}



@app.get('/idoc/<id>')
def redirection_to_idoc(id):
    return bottle.redirect('/#/idoc/'+id)

@app.route('/tmp_static/<filepath:path>')
def server_tmp_static(filepath):
    return bottle.static_file(filepath, root=tmp_imgs_dir)


def cache_img(file_like, basename):
    if not file_like:
        #TODO return link to "broken img"
        return ""
    local_file = os.path.join(tmp_imgs_dir, basename)
    tmp_file = tempfile.mktemp(prefix="idoc_", suffix=".jpg")
    with open(tmp_file, "wb") as lf:
        lf.write(file_like.read())
    shutil.move(tmp_file, local_file)
    return server_tmp_static(os.path.basename(local_file))


#Get the information of one Sleep Monitor
@app.get('/device/<id>/last_img')
@error_decorator
def get_device_last_img(id):
    device = ds.get_device_by_id(id)
    if device.info["recognizer"]["status"] == "running":
        file_like = device.last_image()
        if not file_like:
            raise Exception("No image for %s" % id)

        basename = os.path.join(tmp_imgs_dir, id + "_last_img.jpg")
        return cache_img(file_like, basename)


@app.get('/device/<id>/info')
def info(id):
    return ds.get_device_by_id(id).get_info()

@app.get('/device/<id>/controls/<submodule>/<action>')
def device_actions(id, submodule, action):
    # supported actions:
    return ds.get_device_by_id(id).controls(submodule, action)


@app.post('/device/<id>/settings')
def post_settings(id):
    post_data = bottle.request.body.read()
    return ds.get_device_by_id(id).post(post_data)


@app.get('/device/<id>/settings')
def get_settings(id):
    settings = ds.get_device_by_id(id).settings()
    return settings

@app.get('/device/<id>/get_logs')
def get_logs(id):
    logs = ds.get_device_by_id(id).get_logs()
    return logs

@app.post('/device/<id>/post_logs')
def post_logs(id):
    post_data = bottle.request.body.read() # pylint: disable=no-member
    ds.get_device_by_id(id).post_logs(post_data)


@app.get('/device/<id>/list_paradigms')
def get_paradigms(id):
    paradigms = ds.get_device_by_id(id).get_paradigms()
    return paradigms


@app.post('/device/<id>/load_paradigm')
def load_paradigm(id):
    post_data = bottle.request.body.read() # pylint: disable=no-member
    if post_data is None:
        result = {"status": "failure"}
    else:
        result = ds.get_device_by_id(id).post_paradigm(post_data)

    return result


@app.get('/client/<req>')
@error_decorator
def client_info(req):#, device):
    if req == 'info':

        with os.popen('df %s -h' % RESULTS_DIR) as df:
            disk_free = df.read()

        disk_usage = RESULTS_DIR+" Not Found on disk"

        CARDS = {}
        IPs = []

        CFG.load()

        try:
            disk_usage = disk_free.split("\n")[1].split()

            #the following returns something like this: [['eno1', 'ec:b1:d7:66:2e:3a', '192.168.1.1'], ['enp0s20u12', '74:da:38:49:f8:2a', '155.198.232.206']]
            adapters_list = [ [i, netifaces.ifaddresses(i)[17][0]['addr'], netifaces.ifaddresses(i)[2][0]['addr']] for i in netifaces.interfaces() if 17 in netifaces.ifaddresses(i) and 2 in netifaces.ifaddresses(i) and netifaces.ifaddresses(i)[17][0]['addr'] != '00:00:00:00:00:00' ]
            for ad in adapters_list:
                CARDS [ ad[0] ] = {'MAC' : ad[1], 'IP' : ad[2]}
                IPs.append (ad[2])


            with os.popen('git rev-parse --abbrev-ref HEAD') as df:
                GIT_BRANCH = df.read() or "Not detected"
            #df = subprocess.Popen(['git', 'rev-parse', '--abbrev-ref', 'HEAD'], stdout=subprocess.PIPE)
            #GIT_BRANCH = df.communicate()[0].decode('utf-8')

            with os.popen('git status -s -uno') as df:
                NEEDS_UPDATE = df.read() != ""

            #df = subprocess.Popen(['git', 'status', '-s', '-uno'], stdout=subprocess.PIPE)
            #NEEDS_UPDATE = df.communicate()[0].decode('utf-8') != ""

            with os.popen('systemctl status ethoscope_node.service') as df:
                try:
                    ACTIVE_SINCE = df.read().split("\n")[2]
                except:
                    ACTIVE_SINCE = "Not running through systemd"

        except Exception as e:
            logger.error(e)

        return {'active_since': ACTIVE_SINCE, 'disk_usage': disk_usage, 'IPs' : IPs , 'CARDS': CARDS, 'GIT_BRANCH': GIT_BRANCH, 'NEEDS_UPDATE': NEEDS_UPDATE}

    elif req == 'time':
        return {'time':datetime.datetime.now().isoformat()}

    elif req == 'timestamp':
        return {'timestamp': datetime.datetime.now().timestamp() }

    elif req == 'log':
        with os.popen("journalctl -u ethoscope_node -rb") as log:
            l = log.read()
        return {'log': l}

    elif req == 'daemons':
        #returns active or inactive
        for daemon_name in SYSTEM_DAEMONS.keys():

            with os.popen("systemctl is-active %s" % daemon_name) as df:
                SYSTEM_DAEMONS[daemon_name]['active'] = df.read().strip()
        return SYSTEM_DAEMONS

    elif req == 'folders':
        return CFG.content['folders']

    elif req == 'users':
        return CFG.content['users']

    else:
        raise NotImplementedError()


@app.get('/client-actions/<action>')
def client_actions(action):

    if action == "restart":
        logger.info('User requested a service restart.')
        subprocess.Popen("sleep 1 && systemctl restart IDOC.service", shell=True)
        return ""

def stop(signo=None, _frame=None):
    logger.info("Received signal %s", signo)
    try:
        ds.stop()
        # bottle.abort()
        os._exit(0)
        # sys.exit(0)
    except Exception as error:
        logger.warning(error)
        pass


def main():
    signals = ('TERM', 'HUP', 'INT')
    for sig in signals:
        signal.signal(getattr(signal, 'SIG' + sig), stop)


    SERVER = get_server(PORT)
    bottle.run(app, host='0.0.0.0', port=PORT, debug=DEBUG, server=SERVER)


if __name__ == "__main__":
    main()


__author__ = 'quentin'

import logging
import os

import cv2

from idoc.server.core.base import Base

try:
    from cv2.cv import CV_FOURCC as VideoWriter_fourcc # type: ignore
    from cv2.cv import CV_AA as LINE_AA # type: ignore
except ImportError:
    from cv2 import VideoWriter_fourcc
    from cv2 import LINE_AA

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

from idoc.configuration import IDOCConfiguration
config = IDOCConfiguration()


class BaseDrawer(Base):

    _font = cv2.FONT_HERSHEY_COMPLEX_SMALL
    _black_colour = (0, 0, 0)
    _roi_colour = (0, 255, 0)
    _quality = cv2.IMWRITE_JPEG_QUALITY
    _interaction_colour = (255, 0, 0)
    _standard_colour = (0, 0, 255)

    def __init__(self, *args, video_out=None, draw_frames=True, video_out_fourcc="DIVX", framerate=2, **kwargs):
        """
        A template class to annotate and save the processed frames. It can also save the annotated frames in a video
        file and/or display them in a new window. The :meth:`~ethoscope.drawers.drawers.BaseDrawer._annotate_frame`
        abstract method defines how frames are annotated.

        :param video_out: The path to the output file (.avi)
        :type video_out: str
        :param draw_frames: Whether frames should be displayed on the screen (a new window will be created).
        :type draw_frames: bool
        :param video_out_fourcc: When setting ``video_out``, this defines the codec used to save the output video (see `fourcc <http://www.fourcc.org/codecs.php>`_)
        :type video_out_fourcc: str
        :param framerate: When setting ``video_out``, this defines the output fps. typically, the same as the input fps.
        :type framerate: float
        """
        super().__init__(*args, **kwargs)
        self._settings.update({"framerate": None})

        self.video_out = video_out
        self._draw_frames = draw_frames
        self._video_writers = {}
        self._window_name = "ethoscope_" + str(os.getpid())
        self._video_out_fourcc = video_out_fourcc
        self._settings['framerate'] = framerate
        if draw_frames:
            cv2.namedWindow(self._window_name, cv2.WINDOW_AUTOSIZE)
        self._last_drawn_frame = None
        self.arena = None
        self._frame_count = 0
        self._original_video = None
        self._annotated_video = None


    @property
    def frame_count(self):
        return self._frame_count


    def framerate(self):
        return round(self._settings['framerate'])


    def _annotate_frame(self, img, tracking_units, positions, roi, metadata=None):
        """
        Abstract method defining how frames should be annotated.
        The `img` array, which is passed by reference, is meant to be modified by this method.

        :param img: the frame that was just processed
        :type img: :class:`~numpy.ndarray`
        :param positions: a list of positions resulting from analysis of the frame
        :type positions: list(:class:`~ethoscope.core.data_point.DataPoint`)
        :param tracking_units: the tracking units corresponding to the positions
        :type tracking_units: list(:class:`~ethoscope.core.tracking_unit.TrackingUnit`)
        :return:
        """
        raise NotImplementedError

    @property
    def last_drawn_frame(self):
        return self._last_drawn_frame

    @property
    def last_drawn_path(self):
        return config.content["drawer"]["last_drawn_path"]

    @property
    def last_annot_path(self):
        return config.content["drawer"]["last_annot_path"]

    def write_frames(self):

        cv2.imwrite(self.last_drawn_path, self._last_raw_frame, [int(self._quality), 50])
        cv2.imwrite(self.last_annot_path, self._last_annot_frame, [int(self._quality), 50])

    def write_videos(self, video_root='video.avi'):


        if len(self._video_writers) == 0:

            # TODO More elegant way of deriving the path
            self._original_video = video_root.replace(".avi", "_ORIGINAL.avi")
            self._annotated_video = video_root.replace(".avi", "_ANNOTATED.avi")

            logger.info('Videos will be saved to:')
            logger.info(self._original_video)
            logger.info(self._annotated_video)
            logger.info('With video framerate %s', self.framerate)

            self._video_writers = {

                "ORIGINAL": cv2.VideoWriter(
                    # path to resulting video
                    self._original_video,
                    # codec
                    VideoWriter_fourcc(*self._video_out_fourcc),
                    # framerate
                    self.framerate,
                    # resolution (reverse of first two dims in shape)
                    (self._last_raw_frame.shape[1], self._last_raw_frame.shape[0])
                ),

                "ANNOTATED": cv2.VideoWriter(
                    # path to resulting video
                    self._annotated_video,
                    # codec
                    VideoWriter_fourcc(*self._video_out_fourcc),
                    # framerate
                    self.framerate,
                    # resolution (reverse of first two dims in shape)
                    (self._last_annot_frame.shape[1], self._last_annot_frame.shape[0])
                )
            }

        if len(self._video_writers) == 2:
            self._video_writers["ORIGINAL"].write(self._last_raw_frame)
            self._video_writers["ANNOTATED"].write(self._last_annot_frame)


    def draw(self, img, tracking_units, positions, roi=True, metadata=None):
        """
        Draw results on a frame.

        :param img: the frame that was just processed
        :type img: :class:`~numpy.ndarray`
        :param positions: a list of positions resulting from analysis of the frame by a tracker
        :type positions: list(:class:`~ethoscope.core.data_point.DataPoint`)
        :param tracking_units: the tracking units corresponding to the positions
        :type tracking_units: list(:class:`~ethoscope.core.tracking_unit.TrackingUnit`)
        :return:
        """

        self._last_drawn_frame = img.copy()
        self._last_raw_frame = img.copy()

        img = self._annotate_frame(self._last_drawn_frame, tracking_units, positions, roi, metadata=metadata)
        self._last_annot_frame = img

        if self._draw_frames:
            cv2.imshow(self._window_name, self._last_drawn_frame)
            cv2.waitKey(0)

        if self.video_out is None:
            return self._last_drawn_frame

        self._frame_count += 1
        return self._last_drawn_frame

    def __del__(self):

        if super().stop():
            return

        if self._draw_frames:
            cv2.waitKey(1)
            cv2.destroyAllWindows()
            cv2.waitKey(1)

        for video_writer in self._video_writers.values():
            logger.info('Releasing video writer')
            video_writer.release()

    def close(self):
        # logger.warning("Drawer has drawn %d frames", self._frame_count)
        self.__del__()


class NullDrawer(BaseDrawer):
    def __init__(self):
        """
        A drawer that does nothing (no video writing, no annotation, no display on the screen).

        :return:
        """
        super(NullDrawer, self).__init__(draw_frames=False)

    def _annotate_frame(self, img, tracking_units, positions, roi, metadata=None):
        pass


class DefaultDrawer(BaseDrawer):
    def __init__(self, *args, **kwargs):
        """
        The default drawer. It draws ellipses on the detected objects and polygons around ROIs. When an "interaction"
        see :class:`~ethoscope.stimulators.stimulators.BaseInteractor` happens within a ROI,
        the ellipse is red, blue otherwise.

        :param video_out: The path to the output file (.avi)
        :type video_out: str
        :param draw_frames: Whether frames should be displayed on the screen (a new window will be created).
        :type draw_frames: bool
        """
        super().__init__(*args, **kwargs)

    def _annotate_frame(self, img, tracking_units, positions=None, roi=True, metadata=None):
        if img is None:
            return

        for track_u in tracking_units:

            x, y = track_u.roi.offset
            y += track_u.roi.rectangle[3]/2


            if roi:
                text_pos = (int(x), int(y))
                cv2.putText(img, str(track_u.roi.idx), text_pos, self._font, 1, (255, 255, 0))
                cv2.drawContours(img, [track_u.roi.polygon], -1, self._black_colour, 3, LINE_AA)
                cv2.drawContours(img, [track_u.roi.polygon], -1, self._roi_colour, 1, LINE_AA)

            if positions is not None:
                try:
                    pos_list = positions[track_u.roi.idx]
                except KeyError:
                    continue

                colour = self._standard_colour

                for pos in pos_list:
                    try:
                        if pos["has_interacted"]:
                            colour = self._interaction_colour
                    except KeyError:
                        pass

                    cv2.ellipse(img, ((pos["x"], pos["y"]), (pos["w"], pos["h"]), pos["phi"]), self._black_colour, 3, LINE_AA)
                    cv2.ellipse(img, ((pos["x"], pos["y"]), (pos["w"], pos["h"]), pos["phi"]), colour, 1, LINE_AA)

            if metadata is not None:
                cv2.putText(img, metadata["time"], (50, 30), self._font, 2, (255, 255, 255))
                cv2.putText(img, "%s" % str(metadata["frame"]).zfill(5), (img.shape[1] - 250, 30), self._font, 2, (255, 255, 255))
                cv2.putText(img, "%s" % str(metadata["framerate"]).zfill(2), (img.shape[1] - 100, 30), self._font, 2, (255, 255, 255))


        return img

# TODO Finish this when you have time
# class AnnotationDrawer(BaseDrawer):
#     def __init__(self, video_out= None, draw_frames=False):
#         """
#         The annotation drawer. It draws dots and arrows on the detected objects and polygons around ROIs,
#         instead of an ellipse, so the user can see the contours of the fly while knowing where the tracker
#         thinks the fly is

#         :param video_out: The path to the output file (.avi)
#         :type video_out: str
#         :param draw_frames: Whether frames should be displayed on the screen (a new window will be created).
#         :type draw_frames: bool
#         """
#         super(AnnotationDrawer, self).__init__(video_out=video_out, draw_frames=draw_frames)

#     def _annotate_frame(self,img, tracking_units, positions=None, roi=True):
#         if img is None:
#             return

#         roi_colour = (0, 255,0)
#         for track_u in tracking_units:

#             x,y = track_u.roi.offset
#             y += track_u.roi.rectangle[3]/2
#             black_colour = (0, 0,0)

#             if roi:
#                 cv2.putText(img, str(track_u.roi.idx), (int(x)*5,int(y)), cv2.FONT_HERSHEY_COMPLEX_SMALL, 1, (255,255,0))
#                 cv2.drawContours(img,[track_u.roi.polygon],-1, black_colour, 3, LINE_AA)
#                 cv2.drawContours(img,[track_u.roi.polygon],-1, roi_colour, 1, LINE_AA)

#             if positions is not None:
#                 try:
#                     pos_list = positions[track_u.roi.idx]
#                 except KeyError:
#                     continue

#                 for pos in pos_list:
#                     colour = (0 ,0, 255)
#                     try:
#                         if pos["has_interacted"]:
#                             colour = (255, 0,0)
#                     except KeyError:
#                         pass

#                     # cv2.ellipse(img,((pos["x"],pos["y"]), (pos["w"],pos["h"]), pos["phi"]),black_colour,3, LINE_AA)
#                     # cv2.ellipse(img,((pos["x"],pos["y"]), (pos["w"],pos["h"]), pos["phi"]),colour,1, LINE_AA)
#                     cv2.circle(img, (pos["x"], pos["y"]), 3, black_colour, -1)
#                     cv2.arrowedLine(img, (pos))

#                     distance_raw = round(10**(pos["distance"]/1000), 5)
#                     x,y = self.roi.offset
#                     # x += 0.9*self.roi.rectangle[3]
#                     y += self.roi.rectangle[3]/2

#                     cv2.putText(out, str(distance_raw), (int(x), int(y)), cv2.FONT_HERSHEY_COMPLEX_SMALL, 1, (255, 255, 0))


#         return img



__author__ = 'quentin'

import logging
from ethoscope.stimulators.stimulators import BaseStimulator, HasInteractedVariable
from ethoscope.utils.scheduler import Scheduler
from ethoscope.hardware.interfaces.interfaces import  DefaultInterface
from ethoscope.hardware.interfaces.odour_delivery_device import OdourDelivererInterface, OdourDepriverInterface, OdourDelivererFlushedInterface
from ethoscope.hardware.interfaces.odour_delivery_device import OdourDelivererInterface, OdourDepriverInterface
from . import sleep_depriver_stimulators
import random

class HasChangedSideStimulator(BaseStimulator):
    _HardwareInterfaceClass = DefaultInterface

    def __init__(self, hardware_connection=None, middle_line=0.50):
        """
        class implementing a stimulator that decides whether an animal has change side in its ROI.
        :param hardware_connection: a default hardware interface object
        :param middle_line: the x position defining the line to be crossed (from 0 to 1, relative to ROI)
        :type middle_line: float
        """
        self._middle_line = middle_line
        #self._last_active = 0
        super(HasChangedSideStimulator, self).__init__(hardware_connection)

    def _has_changed_side(self):
        positions = self._tracker.positions

        if len(positions ) <2 :
            return False

        w = float(self._tracker._roi.get_feature_dict()["w"])
        if len(positions[-1]) != 1:
            raise Exception("This stimulator can only work with a single animal per ROI")
        x0 = positions[-1][0]["x"] / w
        xm1 = positions[-2][0]["x"] / w


        if x0 > self._middle_line:
            current_region = 2
        else:
            current_region = 1

        if xm1 > self._middle_line:
            past_region = 2
        else:
            past_region = 1

        if self._tracker._roi == 1:
            logging.warning(str((x0, xm1, current_region, past_region)))

        if current_region != past_region:
            # we return the current region as 1 or 2 if it has changed
            return current_region
        else:
            # no change => 0
            return 0

    def _decide(self):
        has_changed = self._has_changed_side()

        if  has_changed:
            return HasInteractedVariable(True), {}
        return HasInteractedVariable(False), {}

class DynamicOdourDeliverer(HasChangedSideStimulator):
    _description = {"overview": "A stimulator to deliver an odour according to which side the animal of its ROI is in",
                    "arguments": [
                                {"type": "date_range", "name": "date_range",
                                 "description": "A date  and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                                 "default": ""}
                                   ]}

    _HardwareInterfaceClass =  OdourDelivererInterface
    _roi_to_channel = {
            1:1,  2:2,  3:3,  4:4,  5:5,
            6:6, 7:7, 8:8, 9:9, 10:10
        }
    _side_to_pos = {1:1, 2:2 }
    def __init__(self,
                 hardware_connection,
                 date_range=""
                 ):
        """
        A stimulator to control a sleep depriver module

        :param hardware_connection: the sleep depriver module hardware interface
        :type hardware_connection: :class:`~ethoscope.hardawre.interfaces.`
        :return:
        """

        self._t0 = None
        self._scheduler = Scheduler(date_range)
        super(DynamicOdourDeliverer, self).__init__(hardware_connection)



    def _decide(self):
        roi_id= self._tracker._roi.idx
        try:
            channel = self._roi_to_channel[roi_id]
        except KeyError:
            return HasInteractedVariable(False), {}

        if self._scheduler.check_time_range() is False:
            return HasInteractedVariable(False), {}

        has_changed_side = self._has_changed_side()

        if has_changed_side == 0:
            return HasInteractedVariable(False), {}
        pos = self._side_to_pos[has_changed_side]
        return HasInteractedVariable(pos), {"channel":channel, "pos" : self._side_to_pos[has_changed_side]}

class DynamicOdourSleepDepriver(sleep_depriver_stimulators.SleepDepStimulator):
    _description = {
        "overview": "An stimulator to sleep deprive an animal using servo motor. See http://todo/fixme.html",
        "arguments": [
            {"type": "number", "min": 0.0, "max": 1.0, "step": 0.0001, "name": "velocity_correction_coef",
             "description": "Velocity correction coef", "default": 3.0e-3 / 2},
            {"type": "number", "min": 2.0, "max": 10.0, "step": 0.5, "name": "stimulus_duration",
             "description": "How long to send the puff of odour for", "default": 5.0},
            {"type": "number", "min": 1, "max": 3600 * 12, "step": 1, "name": "min_inactive_time",
             "description": "The minimal time after which an inactive animal is awaken", "default": 120},
            {"type": "date_range", "name": "date_range",
             "description": "A date  and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
             "default": ""}
        ]}

    _HardwareInterfaceClass =  OdourDepriverInterface
    _roi_to_channel = {
            1:1,  2:2,  3:3,  4:4,  5:5,
            6:6, 7:7, 8:8, 9:9, 10:10
        }

    def __init__(self,
                 hardware_connection,
                 velocity_correction_coef=3.0e-3 / 2,
                 min_inactive_time=120,  # s
                 stimulus_duration=5,  #s
                 date_range=""
                 ):
        """
        A stimulator to control an odour sleep depriver module.

        :param hardware_connection: the sleep depriver module hardware interface
        :type hardware_connection: :class:`~ethoscope.hardawre.interfaces.sleep_depriver_interface.SleepDepriverInterface`
        :param velocity_correction_coef: correct velocity by this coefficient to make it fps-inveriant. 1 => walking
        :type velocity_correction_coef: float
        :param stimulus_duration: how long the odour delivery takes place for
        :type stimulus_duration: float
        :param min_inactive_time: the minimal time without motion after which an animal should be disturbed (in seconds)
        :type min_inactive_time: float
        :return:
        """
        self._stimulus_duration = stimulus_duration
        super(DynamicOdourSleepDepriver, self).__init__(hardware_connection, velocity_correction_coef, min_inactive_time, date_range)

    def _decide(self):
        decide, args = super(DynamicOdourSleepDepriver, self)._decide()
        args["stimulus_duration"] = self._stimulus_duration

        return decide, args



class MiddleCrossingOdourStimulator(sleep_depriver_stimulators.MiddleCrossingStimulator):
    _description = {"overview": "A stimulator to send odour to an animal as it crosses the midline",
                    "arguments": [
                        {"type": "number", "min": 0.0, "max": 1.0, "step": 0.01, "name": "p",
                         "description": "the probability to move the tube when a beam cross was detected",
                         "default": 1.0},
                        {"type": "number", "min": 0.0, "max": 300, "step": 1, "name": "refractory_period",
                         "description": "cannot send two stimuli if they are not separated from, at least, this duration",
                         "default": 120},
                        {"type": "number", "min": 2.0, "max": 10.0, "step": 0.5, "name": "stimulus_duration",
                         "description": "How long to send the puff of odour for", "default": 5.0},
                        {"type": "date_range", "name": "date_range",
                         "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                         "default": ""}
                    ]}

    _HardwareInterfaceClass = OdourDepriverInterface

    _roi_to_channel = {
            1:1,  2:2,  3:3,  4:4,  5:5,
            6:6, 7:7, 8:8, 9:9, 10:10
        }



    def __init__(self,
                 hardware_connection,
                 p=1.0,
                 refractory_period = 300,
                 stimulus_duration = 5,
                 date_range=""
                 ):

        super(MiddleCrossingOdourStimulator, self).__init__(hardware_connection, p=p, date_range=date_range)
        self._refractory_period = refractory_period
        self._stimulus_duration = stimulus_duration


    def _decide(self):
        decide, args = super(MiddleCrossingOdourStimulator, self)._decide()
        args["stimulus_duration"] = self._stimulus_duration

        return decide, args

class MiddleCrossingOdourStimulatorFlushed(MiddleCrossingOdourStimulator):
    _description = {"overview": "A stimulator to send odour to an animal as it crosses the midline, and then flush it",
                    "arguments": [
                        {"type": "number", "min": 0.0, "max": 1.0, "step": 0.01, "name": "p",
                         "description": "the probability to move the tube when a beam cross was detected",
                         "default": 1.0},
                        {"type": "number", "min": 0.0, "max": 300, "step": 1, "name": "refractory_period",
                         "description": "cannot send two stimuli if they are not separated from, at least, this duration",
                         "default": 120},
                        {"type": "number", "min": 2.0, "max": 10.0, "step": 0.5, "name": "stimulus_duration",
                         "description": "How long to send the puff of odour for", "default": 5.0},
                        {"type": "number", "min": 2.0, "max": 60.0, "step": 0.5, "name": "flush_duration",
                         "description": "How long to flush odour for", "default": 10.0},
                        {"type": "date_range", "name": "date_range",
                         "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                         "default": ""}
                    ]}

    _HardwareInterfaceClass = OdourDelivererFlushedInterface
    _roi_to_channel = {
            1:1,  2:2,  3:3,  4:4,  5:5,
            6:6, 7:7, 8:8, 9:9, 10:10
        }

    def __init__(self,
                 hardware_connection,
                 p=1.0,
                 refractory_period = 300,
                 stimulus_duration = 5,
                 flush_duration=10,
                 date_range=""
                 ):

        super(MiddleCrossingOdourStimulator, self).__init__(hardware_connection, p=p, date_range=date_range)
        self._refractory_period = refractory_period
        self._stimulus_duration = stimulus_duration
        self._flush_duration = flush_duration


    def _decide(self):
        decide, args = super(MiddleCrossingOdourStimulator, self)._decide()
        args["stimulus_duration"] = self._stimulus_duration
        args["flush_duration"] = self._flush_duration

        return decide, args


from ethoscope.stimulators.sleep_depriver_stimulators import MiddleCrossingStimulator
from ethoscope.hardware.interfaces.optomotor import OptoMotor


class OptoMidlineCrossStimulator (MiddleCrossingStimulator):
    _description = {"overview": "A stimulator to shine light when animals cross the midline",
                    "arguments": [
                                    {"type": "number", "min": 0.0, "max": 1.0, "step":0.01, "name": "p", "description": "the probability to move the tube when a beam cross was detected","default":1.0},
                                    {"type": "date_range", "name": "date_range",
                                     "description": "A date and time range in which the device will perform (see http://tinyurl.com/jv7k826)",
                                     "default": ""}
                                   ]}
    _HardwareInterfaceClass = OptoMotor
    _roi_to_channel = {
        1: 1,
        3: 3,
        5: 5,
        7: 7,
        9: 9,
        12: 23,
        14: 21,
        16: 19,
        18: 17,
        20: 15
    }




__author__ = 'quentin'


class BaseIntVariable(int):
    r"""
    Template class for defining arbitrary variable types.
    Each class derived from this one should at least define the three following attributes:

    * `sql_data_type`, The MySQL data type. Allows to use minimal space to save data points.
    * `header_name`, The name of this variable.
        Used as the column name in the result table, so it must be unique.
    * `functional_type`, A keyword defining what type of variable this is.
        For instance "distance", "angle" or "proba".
        Allows specific post-processing per functional type.
    """

    sql_data_type = "SMALLINT"
    header_name = None
    functional_type = None # {distance, angle, bool, confidence,...}

    def __new__(cls, value):
        if cls.functional_type is None:
            raise NotImplementedError(
                "Variables must have a functional data type such as \
                    'distance', 'angle', 'bool', 'confidence'"
            )
        if cls.sql_data_type is None:
            raise NotImplementedError("Variables must have an SQL data type such as INT")
        if cls.header_name is None:
            raise NotImplementedError("Variables must have a header name")
        return  super(BaseIntVariable, cls).__new__(cls, value)


class BaseBoolVariable(BaseIntVariable):
    """
    Abstract type encoding boolean values.
    Internally stored as int as bool type cannot be derived.
    """
    functional_type = "bool"
    sql_data_type = "BOOLEAN"

class IsInferredVariable(BaseBoolVariable):
    """
    Type encoding whether a data point is inferred (from past values)
    or observed; 0 or 1, respectively.
    """
    header_name = "is_inferred"


class PhiVariable(BaseIntVariable):
    """
    Type encoding the angle of a detected object, in degrees.
    """
    header_name = "phi"
    functional_type = "angle"

class Label(BaseIntVariable):
    """
    Type encoding a discrete label when several objects, in the same ROI, are detected.
    """
    header_name = "label"
    functional_type = "label"


class BaseDistanceIntVar(BaseIntVariable):
    """
    Abstract type encoding variables representing distances.
    """
    functional_type = "distance"

class mLogLik(BaseIntVariable):
    """
    Type representing a log likelihood.
    It should be multiplied by 1000 to be stored as an int.
    """
    header_name = "mlog_L_x1000"
    functional_type = "proba"


class XYDistance(BaseIntVariable):
    """
    Type storing distance moved between two consecutive observations.
    Log10 x 1000 is used so that floating point distance is stored as an int.
    """
    header_name = "xy_dist_log10x1000"
    functional_type = "relative_distance_1e6"


class WidthVariable(BaseDistanceIntVar):
    """
    Type storing the width of a detected object.
    """
    header_name = "w"

class HeightVariable(BaseDistanceIntVar):
    """
    Type storing the height of a detected object.
    """
    header_name = "h"

class BaseRelativeVariable(BaseDistanceIntVar):
    """
    Abstract type encoding distance variables that can be expressed relatively to an origin.
    They converted to absolute using information form the ROI.
    """
    def to_absolute(self, roi):
        """
        Converts a positional variable from a relative (to the top left of a ROI)
        to an absolute (e.i. top left of the parent image).

        :param roi: a region of interest
        :type roi: :class:`~ethoscope.rois.roi_builders.ROI`.
        :return: A new variable
        :rtype: :class:`~ethoscope.core.variable.BaseRelativeVariable`
        """
        return self._get_absolute_value(roi)
    def _get_absolute_value(self, roi):
        raise NotImplementedError(
            "Relative variable must implement a `get_absolute_value()` method"
        )

class XPosVariable(BaseRelativeVariable):
    """
    Type storing the X position of a detected object.
    """
    header_name = "x"
    def _get_absolute_value(self, roi):
        out = int(self)

        x_offset, _ = roi.offset
        out += x_offset
        return XPosVariable(out)

class YPosVariable(BaseRelativeVariable):
    """
    Type storing the Y position of a detected object.
    """
    header_name = "y"
    def _get_absolute_value(self, roi):
        out = int(self)
        _, y_offset = roi.offset
        out += y_offset
        return YPosVariable(out)

class FrameCountVariable(BaseIntVariable):
    """
    Type storing a frame count
    """

    header_name = "frame_count"
    functional_type = "count"
    sql_data_type = "INT"



r"""
Provide abstract classes with functionality used in:
idoc.server.core.control_thread.ControlThread
idoc.server.controllers.controllers.Controller
idoc.server.programmers.programmers.Programmer
idoc.server.io.cameras.StandardStream
idoc.server.io.cameras.PylonStream
idoc.server.io.cameras.WebCamStream
idoc.server.io.saver.SimpleSaver
idoc.server.trackers.trackers.SimpleTracker
"""

import datetime
import logging
from threading import Thread, Event
import traceback

from idoc.helpers import hours_minutes_seconds, iso_format, MachineDatetime
from idoc.configuration import IDOCConfiguration
from idoc.debug import IDOCException

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class Settings:
    r"""
    Communication bridge between modules in IDOC.
    Each module needs to define a dictionary self._settings to be updated
    and a dictioanry of submodules in self._submodules to communicate with.
    Submodules are notified with the send() and their updates are received with receive()
    """

    def __init__(self, *args, **kwargs):
        self._settings = {}
        self._submodules = {}
        super().__init__(*args, **kwargs)


    def _add_submodule(self, name, submodule):
        self._submodules[name] = submodule
        self._settings[name] = submodule.settings

    @property
    def settings(self):
        r"""
        _settings will only ever contain the keys declared in the
        constructor of the classes inheriting from Settings,
        AND also the _settings of instances listed in _submodules.
        """
        return self._settings

    @settings.setter
    def settings(self, settings):
        r"""
        Update the existing settings without extending them.
        """
        # only keep settings that already existed
        new_settings = {k: settings[k] for k in self._settings if k in settings}
        # only update those that are different
        for k in self._settings:
            if k in new_settings:
                if self._settings[k] != new_settings[k]:
                    self._settings[k] = new_settings[k]

        # self._settings.update(new_settings)

        # notify submodules
        self.send()
        # actually apply the changes
        self.update()

        logger.warning("Module %s with settings %s", self.__class__.__name__, self._settings)

    @settings.getter
    def settings(self):
        r"""
        Return updated settings of current class as well as all other
        instances listed in self._submodules.
        """
        # ask the dependent modules
        self.receive()
        self.update()
        return self._settings


    def send_recursive(self, module, settings):
        """
        Take a module and a settings dictionary
        For every submodule in the module and every setting
        If the setting matches the name of a submodule, repeat recursively
        otherwise, the setting is an actual setting and we need to set it
        """
        # TODO Refactor this to make it more readable

        # For each submodule of the passed module
        for name, submodule in module._submodules.items():
            # And for each setting
            for k in self._settings:
                # If the setting is a dictionary for a submodule of the submodule
                if submodule is not None:
                    if k in submodule._submodules:
                        # pick just those settings and call send_recursive again only with the current submodule
                        local_settings = settings[k]
                        self.send_recursive(submodule, local_settings)
                    else:
                        if k in settings and k in submodule._settings:
                            submodule._settings[k] = settings[k]

                    submodule.update()


    def send(self):
        r"""
        Send to depending modules the new settings sent by the user.
        """
        for name, submodule in self._submodules.items():
            if submodule is None:
                continue

            logger.warning("%s sending to %s", self.__class__.__name__, submodule.__class__.__name__)
            logger.warning(self._settings)
            self.send_recursive(submodule, self._settings)

    def receive(self):
        r"""
        Receive from depending modules settings that the user could change
        """
        for name, submodule in self._submodules.items():
            if submodule is None:
                continue

            logger.debug(
                "%s receiving from %s",
                self.__class__.__name__, submodule.__class__.__name__
            )
            logger.debug(submodule.settings)
            self._settings[name] = submodule.settings

    def update(self):
        """
        Update the properties using the values in the settings
        """
        for k in self._settings:
            if k in self._submodules:
                self.send_recursive(self._submodules[k], {k: self._settings[k]})
            else:
                try:
                    setattr(self, k, self._settings[k])
                except Exception as error:
                    logger.warning(f"{self} - Can't set {k} to {self._settings[k]}")
                    logger.warning(error)
                    logger.warning(traceback.print_exc())
                



class Status:
    """
    Report useful data and methods to learn and change the state of a class.
    Three state flags are available: ready, running and stopped.

    * reset()
    * run()
    * stop()
    * status
    * elapsed_time
    * elapsed_seconds
    * start_datetime
    * time_zero

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.running = False
        self.ready = False
        self.stopped = False
        self._time_zero = None
        self._time_init = MachineDatetime.now()
        self._start_datetime = None
        self.experiment_start = None

        self._status = "idle"

        self._info = {}
        self._info.update({
            "status": self._status,
            "elapsed_time": self.elapsed_time
        })

        self._user_data = None


    def reset(self):
        """
        Set all state flags to False i.e. like at __init__
        """
        self.running = False
        self.ready = False
        self.stopped = False

    @property
    def elapsed_time(self):
        """
        Compute the number of seconds since running was set to True
        Return it in format HH:MM:SS.
        """

        if self._time_zero is None:
            return None

        time_diff = iso_format(
            hours_minutes_seconds(
                (datetime.datetime.now() - self._time_zero).seconds
            )
        )
        return time_diff

    @property
    def init_elapsed_seconds(self):
        if self._time_init is None:
            return None

        time_diff = (datetime.datetime.now() - self._time_init).total_seconds()
        return time_diff

    @property
    def elapsed_seconds(self):
        """
        Compute the number of seconds since running was set to True
        Return a float
        """
        if self._time_zero is None:
            return None

        time_diff = (datetime.datetime.now() - self._time_zero).total_seconds()
        return time_diff

    @property
    def experiment_elapsed_seconds(self):

        if self.experiment_start is None:
            return None

        time_diff = (datetime.datetime.now() - self.experiment_start).total_seconds()
        return time_diff

    @property
    def start_datetime(self):
        return self._start_datetime

    @start_datetime.setter
    def start_datetime(self, start_datetime):
        self._start_datetime = start_datetime


    @property
    def time_zero(self):
        """
        Return a datetime.datetime object storing the timestamp
        when running was set to True.
        """

        try:
            return self._time_zero
        except KeyError:
            return None

    @time_zero.setter
    def time_zero(self, time_zero):
        self._time_zero = time_zero

    @property
    def status(self):
        return self._status

    @status.getter
    def status(self):
        """
        Return
        - idle if the the thread has not started yet
        - running if the thread has started and has not been stopped
        - stopped if it has been stopped
        """
        if not self.running:
            self._status = "idle"
            status = self._status

        elif self.ready and not self.running:
            self._status = 'ready'
            status = self._status

        elif self.running and not self.stopped:
            self._status = "running"
            status = self._status

        elif self.stopped:
            self._status = "stopped"
            status = self._status

        else:
            logger.debug('%s status undefined', self.__class__.__name__)
            status = 'undefined'

        self._status = status
        return self._status

    def run(self):
        """
        Set running to True
        Populate _time_zero (datetime.datetime) and _start_datetime (str)
        Should be called from a subclass run method.
        """
        if self.running is True:
            logger.warning("%s is already running. Skipping", self.__class__.__name__)
            return

        self.running = True
        self._set_start_datetime()


    def _set_start_datetime(self):
        """
        Store a timestamp of the moment when running was set to True
        in type str with format YYYY-MM-DD_HH:MM:SS
        """
        start_date_time = MachineDatetime.now()
        self._start_datetime = start_date_time.machineformat()
        self._time_zero = start_date_time


    def stop(self):
        """
        Set the stopped flagged to True.
        All concurrently running methods should listen to this flag
        and react accordingly, tipically interrupting execution.
        Should be called from a subclass run method.
        """
        if self.stopped is True:
            logger.warning("%s is already stopped. Skipping", self.__class__.__name__)
            return True

        self.stopped = True
        return False

class StatusThread(Status, Thread):
    r"""
    Make use of threading.Event instances to report status of subclasses
    and alter their behaviour accordingly.
    """

    def __init__(self, *args, **kwargs):
        self._start_event = Event()
        self._ready_event = Event()
        self._end_event = Event()
        super().__init__(*args, **kwargs)

    def reset(self):

        self.stop()
        self._start_event.clear()
        self._ready_event.clear()
        self._end_event.clear()


    def run(self):
        super().run()
        Thread.run(self)


    @property
    def running(self):
        return self._start_event.is_set()

    @running.setter
    def running(self, value):
        if value is True:
            self._start_event.set()

    @property
    def ready(self):
        return self._ready_event.is_set()

    @ready.setter
    def ready(self, value):
        if value is True:
            self._ready_event.set()

    @property
    def stopped(self):
        return self._end_event.is_set()

    @stopped.setter
    def stopped(self, value):
        if value is True:
            self._end_event.set()

class DescribedObject:
    r"""
    An object that contains a ``description`` attribute.
    This is used to parse user option for the web interface.
    This way, users can send option to the different objects used.
    ``description`` is a dictionary with the fields "overview" and "arguments".
    "overview" is simply a string. "arguments" is a list of dictionaries. Each has the field:

     * name: The name of the argument as it is in "__init__"
     * description: "A user friendly description of the argument"
     * type: "number", "datetime", "daterange" and "string".
     * min, max and step: only for type "number",
       defines the accepted limits of the arguments
       as well as the increment in the user interface
     * default: the default value

    Each argument must match a argument in `__init__`.
    """
    _description = None

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)

    @property
    def description(self): # pylint: disable=missing-function-docstring
        return self._description


class Base(Settings, StatusThread):
    r"""
    Provide the functionality in Settings and StatusThread simultaneously,
    which is the standard in IDOC.
    """
    def __init__(self, *args, **kwargs): # pylint: disable=useless-super-delegation
        super().__init__(*args, **kwargs)
        self._validate()


    def _validate(self):
        """
        Make sure the _valid_actions have a corresponding method
        """

        valid_actions = getattr(self, "_valid_actions", None)
        if valid_actions is not None:
            for action in valid_actions:
                if not getattr(self, action, False):
                    raise IDOCException("Action %s of class %s does not have a matching method" % (action, self.__class__.__name__))



import os.path
from setuptools import setup, find_packages
with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION="2.1.2"
PKG_NAME="idoc"
with open(os.path.join(PKG_NAME, "__init__.py"), "w") as filehandle:
    filehandle.write(f"__version__=\"{VERSION}\"\n")


setup(
    name=PKG_NAME,
    version=VERSION,
    author="Antonio Ortega",
    author_email="antonio.ortega@kuleuven.be",
    description="Individual Drosophila Optogenetics Conditioner",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shaliulab/idoc",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    include_package_data=True,
    install_requires = [
        "pypylon",
        "opencv-python>=4.1.1",
        "pyserial>=3.4",
        "scipy",
        "numpy",
        "pandas",
        "pyfirmata",
        "pyaml",
        "coloredlogs",
        "bottle",
        "cheroot",
        "netifaces",
        "requests",
    ],
    entry_points={
        "console_scripts": [
            "idoc_server=idoc.server.bin.server:main",
            "idoc_client=idoc.client.bin.client:main",
            "idoc_cli=idoc.client.bin.cli:main",
            "idoc_offline=idoc.server.bin.offline:offline",
            "idoc_config=idoc.configuration:integrity_check",
            "idoc_batch=idoc.server.bin.control:main",
            "idoc_postinstall=idoc.postinstall:main",
        ]
    },
    python_requires='>=3.8.10',
)



import cv2
import numpy as np
import logging
logging.basicConfig(level=logging.INFO)


def center2rect(center, height, left, right):

    half_h = height // 2

    tl = center + np.array([-left, -half_h])
    tr = center + np.array([+right, -half_h])
    br = center + np.array([+right, +half_h])
    bl = center + np.array([-left, +half_h])

    ct = np.array([tl, tr, br, bl], dtype=np.int32)
    return ct

def contour_center(cnt):
    M = cv2.moments(cnt)
    cx = int(M['m10']/M['m00'])
    cy = int(M['m01']/M['m00'])
    return (cx, cy)

def rotate_contour(cnt, angle, center_of_mass=None):
    if angle != 0.0:
        if center_of_mass is None:
            center_of_mass = contour_center(cnt)

        M = cv2.getRotationMatrix2D(center_of_mass, angle, 1.0)

        cnt_shape = list(cnt.shape)
        coord_slot = cnt_shape.index(2)
        cnt_shape[coord_slot] = 1

        cnt_z = np.append(cnt, np.zeros(tuple(cnt_shape)), axis=coord_slot)
        cnt_rot = np.round(np.dot(M, cnt_z.T).T).astype(np.int32)
        return cnt_rot
    else:
        return cnt

def move_contour(cnt, pixel=0, axis=1):
    if pixel == 0:
        return cnt
    else:
        x = pixel * (axis == 0)
        y = pixel * (axis == 1)
        cnt_moved = np.array([[pt[0] + x, pt[1] + y] for pt in cnt])
        return cnt_moved


def refine_contour(cnt, grey, rotate=True, move=True):

    if rotate:
        max_angle = 0.0
        learning_rate = 0.01
        center_of_mass = contour_center(cnt)

        cnt_rot = rotate_contour(cnt, +learning_rate, center_of_mass)
        mean_pos = contour_mean_intensity(grey, cnt_rot)
        cnt_rot = rotate_contour(cnt, -learning_rate, center_of_mass)
        mean_neg = contour_mean_intensity(grey, cnt_rot)
        gradient = np.array([-1,1])[np.argmin(np.array([mean_neg, mean_pos]))]

        original_val = contour_mean_intensity(grey, cnt)
        max_val = original_val
        for angle in np.arange(-.25, .25, learning_rate):

        # while not min_found and n_iters < 100:
            inner_cnt_rot = rotate_contour(cnt, angle, center_of_mass)
            val = contour_mean_intensity(grey, inner_cnt_rot)
            if val > max_val:
                max_val = val
                max_angle = angle

        cnt_rot = rotate_contour(cnt, max_angle, center_of_mass)
        if max_angle != 0:
            cv2.drawContours(grey, [inner_cnt_rot], -1, (255, 0, 255), 2)

        cv2.drawContours(grey, [cnt_rot], -1, (255, 255, 0), 2)

        cnt_rot.reshape((1, *cnt_rot.shape))

    else:
        cnt_rot = cnt
        max_angle = 0


    if move:
        original_val = contour_mean_intensity(grey, cnt_rot)
        max_val = original_val
        max_pixel = 0
        # print("Running moving algorithm for ONE contour")
        for pixel in np.arange(-10, 10, 1):
            inner_cnt_moved = move_contour(cnt=cnt_rot, pixel=pixel)
            val = contour_mean_intensity(grey, inner_cnt_moved)

            if val > max_val:
                max_val = val
                max_pixel = pixel

        optim_vertical = move_contour(cnt=cnt_rot, pixel=max_pixel)
        original_val = max_val
        max_pixel = 0
        for pixel in np.arange(-100, 100, 10):
            inner_cnt_moved = move_contour(cnt=optim_vertical, pixel=pixel,axis=0)
            val = contour_mean_intensity(grey, inner_cnt_moved)


            if val > max_val:
                max_val = val
                max_pixel = pixel

        final_contour = move_contour(cnt=optim_vertical, pixel=int(max_pixel*0.8),axis=0)




    else:
        final_contour = cnt_rot
        max_pixel = 0
    return final_contour, grey, max_angle, max_pixel

def contour_mean_intensity(grey, cnt):
    """
    A np.array with shape ndotsx2
    """

    grey = cv2.cvtColor(grey, cv2.COLOR_BGR2GRAY)

    mask = np.zeros_like(grey, dtype=np.uint8)
    mask = cv2.drawContours(mask, [cnt], -1, 255, -1)
    mean = cv2.mean(grey, mask=mask)[0]

    return mean

def place_dots(grey, pts, color=255):

    # logging.warning("Placement")
    # logging.warning(pts)
    for pt in pts:
        pt = tuple([int(e) for e in pt])
        cv2.circle(grey, pt, 2, color, -1)
    return grey


def pull_contour_h(cnt, point, side="left"):
    max_dist = -10

    sign = -1 if side == "left" else + 1
    funcs = {"left": np.max, "right": np.min}
    distances = np.array([point[0] - corner[0] for corner in cnt])
    dist = funcs[side](distances)
    diff = int(dist - sign*max_dist)
    cnt = np.array([[corner[0] + diff, corner[1]] for corner in cnt])
    return cnt


def find_quadrant(shape, center):
    # todo dont hardcode this
    left = center[0] < (shape[1] / 2)
    top = center[1] > (shape[0] / 2)
    return (left, top)
